
import unittest
import logging
from amscrot.client import Client
from amscrot.util import utils

class TestClientModule(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.INFO)
        self.session_name = "test_client_module_session"

    def test_client_lifecycle_dummy_provider(self):
        client = Client()

        # Add a dummy provider
        provider = client.add_provider(
            label="my_dummy_provider",
            type="dummy",
            config={} 
        )

        # Create Session
        session = client.create_session(self.session_name)

        # Add a dummy resource
        session.add_node(
            label="my_dummy_node",
            provider=provider,
            count=1,
            image="default_image",
            flavor="default_flavor"
        )

        # 1. Plan
        print("\n--- Running Plan ---")
        created, deleted = session.plan()
        self.assertEqual(created, 1)
        self.assertEqual(deleted, 0)

        # 2. Apply
        print("\n--- Running Apply ---")
        status = session.apply()
        self.assertEqual(status, 0) # 0 means success in amscrot apply

        # 3. Show
        print("\n--- Running Show ---")
        session.show()

        # 4. Destroy
        print("\n--- Running Destroy ---")
        status = session.destroy()
        self.assertEqual(status, 0) # 0 means success in amscrot destroy

if __name__ == "__main__":
    unittest.main()



"""Stitch a link through the Utah network to the FABRIC port on scidmz-cl
"""

# Import the Portal object.
import geni.portal as portal
# Import the ProtoGENI library.
import geni.rspec.pg as pg
# Emulab extensions
import geni.rspec.emulab
from ipaddress import IPv4Network, IPv4Address


NODE_MIN=0
NODE_MAX=10
VLAN_MIN=1000
VLAN_MAX=4095

# Create a portal context.
pc = portal.Context()

#
# List of Cloudlab clusters that have Fabric support.
#
clusters = [
    ('urn:publicid:IDN+utah.cloudlab.us+authority+cm', 'Utah'),
    ('urn:publicid:IDN+clemson.cloudlab.us+authority+cm', 'Clemson')]

pc.defineParameter("cluster", "Select Cluster",
                   portal.ParameterType.STRING,
                   clusters[0], clusters,
                   longDescription="Select a cluster")

# parameterize the vlan to use
portal.context.defineParameter("vlan", "VLAN ID", portal.ParameterType.INTEGER, 3100)
portal.context.defineParameter("ip_subnet", "IP_SUBNET", portal.ParameterType.STRING, "192.168.1.0/24")
portal.context.defineParameter("ip_start", "IP_START", portal.ParameterType.STRING, "192.168.1.2")
portal.context.defineParameter("node_count", "NODE_COUNT", portal.ParameterType.INTEGER, NODE_MIN)
params = portal.context.bindParameters()

# Check parameter validity.
if params.vlan < VLAN_MIN or params.vlan > VLAN_MAX:
    portal.context.reportError( portal.ParameterError( "VLAN ID must be in the range {}-{}".format(VLAN_MIN, VLAN_MAX) ) )

try:
    subnet = IPv4Network(unicode(params.ip_subnet))
    ip_start = int(IPv4Address(unicode(params.ip_start)))
except Exception as e:
    raise Exception("Bad subnet or ipstart")

if params.node_count < NODE_MIN or params.node_count > NODE_MAX:
    portal.context.reportError( portal.ParameterError( "Node count must be between {} and {} inclusive".format(NODE_MIN, NODE_MAX) ) )

# Create a Request object to start building the RSpec.
request = pc.makeRequestRSpec()

interfaces = list()

# Request nodes at one of the Utah clusters (Cloudlab Utah, Emulab, APT)

for i in range(0, params.node_count):
    node = request.RawPC("node{}".format(i))
    node.disk_image = "urn:publicid:IDN+emulab.net+image+emulab-ops//UBUNTU22-64-STD"
    # node.hardware_type = "m510"

    if (params.cluster == 'urn:publicid:IDN+clemson.cloudlab.us+authority+cm'):
        node.Site("utah")
    else:
        node.Site("clemson")

    iface = node.addInterface("if0")
    # Must specify the IPv4 address on all stitched links
    iface.addAddress(pg.IPv4Address(str(IPv4Address(ip_start)), str(subnet.netmask)))
    ip_start += 1
    interfaces.append(iface)
    node.component_manager_id = params.cluster

# Request a special node that maps to the scidmz FABRIC port
fabric = request.Node("stitch-node", "emulab-connect")
fabric.Site("stitch")

# Magic.
fabric.component_id = "interconnect-fabric"
# XXX special handling for stitch fabric component_manager_id
if (params.cluster == 'urn:publicid:IDN+clemson.cloudlab.us+authority+cm'):
    fabric.component_manager_id = params.cluster
else:
    fabric.component_manager_id = "urn:publicid:IDN+stitch.geniracks.net+authority+cm"

fabric.exclusive = False
siface = fabric.addInterface("if0")
# Specify the IPv4 address for stitch node
siface.addAddress(pg.IPv4Address(str(IPv4Address(ip_start)), str(subnet.netmask)))
interfaces.append(siface)

# Make a LAN
if params.node_count == 1:
    lan = request.Link("link", "vlan")
else:
    lan = request.LAN()

for iface in interfaces:
    lan.addInterface(iface)

# Request one of the allowed tags
lan.setVlanTag(params.vlan)

# Many nodes have a single physical experimental interface, so use
# link multiplexing to make sure it maps to any node.
lan.link_multiplexing = True;

# Use best effort on stitched links.
lan.best_effort = True;

# Print the RSpec to the enclosing page.
pc.printRequestRSpec(request)


from ipaddress import IPv4Network, IPv6Network

DEFAULT_RENEWAL_IN_DAYS = 14
DEFAULT_OC_HOST = 'orchestrator.fabric-testbed.net'
DEFAULT_CM_HOST = 'cm.fabric-testbed.net'
DEFAULT_BASTION_HOST = 'bastion.fabric-testbed.net'

FABRIC_OC_HOST = "oc-host"
FABRIC_CM_HOST = "cm-host"
FABRIC_TOKEN_LOCATION = "token-location"
FABRIC_BASTION_HOST = "bastion-host"
FABRIC_BASTION_USER_NAME = "bastion-user-name"
FABRIC_BASTION_KEY_LOCATION = "bastion-key-location"

FABRIC_RANDOM = "FABRIC.RANDOM"
FABRIC_PROJECT_ID = "project_id"

FABRIC_SLICE_PRIVATE_KEY_LOCATION = "slice-private-key-location"
FABRIC_SLICE_PUBLIC_KEY_LOCATION = "slice-public-key-location"

FAB_CONF_ATTRS = [
    FABRIC_TOKEN_LOCATION,
    FABRIC_BASTION_USER_NAME,
    FABRIC_BASTION_KEY_LOCATION,
    FABRIC_SLICE_PRIVATE_KEY_LOCATION,
    FABRIC_SLICE_PUBLIC_KEY_LOCATION
]

FABRIC_IPV4_NET_IFACE_NAME = "v4_net_iface"
FABRIC_IPV6_NET_IFACE_NAME = "v6_net_iface"
FABRIC_IPV4_NET_NAME = "v4_net"
FABRIC_IPV4EXT_NET_NAME = "v4ext_net"
FABRIC_IPV6_NET_NAME = "v6_net"
FABRIC_IPV6EXT_NET_NAME = "v6ext_net"
FABRIC_STITCH_NET_NAME = "stitch_net"
FABRIC_STITCH_NET_IFACE_NAME = "stitch_net_iface"

# RFC1918 FABRIC subnet
FABRIC_PRIVATE_IPV4_SUBNET = IPv4Network('10.128.0.0/10')
# FABRIC public ipv4 subnet
FABRIC_PUBLIC_IPV4_SUBNET = IPv4Network('23.134.232.0/22')
# FABRIC public ipv6 subnet
FABRIC_PUBLIC_IPV6_SUBNET = IPv6Network('2602:FCFB::0/40')

FABNET_IPV4_PREFIX = "FABNET_IPv4"
FABNET_IPV4EXT_PREFIX = "FABNET_IPv4Ext"
FABNET_IPV6_PREFIX = "FABNET_IPv6"
FABNET_IPV6EXT_PREFIX = "FABNET_IPv6Ext"

FABRIC_SLEEP_AFTER_SUBMIT_OK = 120  # In seconds

PATCH_FOR_TOKENS = True

SITES = "sites"



from amscrot.policy.policy_helper import *


def test_policy_with_site_and_no_site():
    yaml_str = '''
fabric:
  stitch-port:
      - name: AAA
        member-of:
          - UTAH
        profile: prof1
        preference: 10
        site: sitea
      - name: BBB
        member-of:
          - UTAH
        profile: prof2
        preference: 100
        site: siteb
  group:
      - name: UTAH
        consumer-for:
          - cloudlab        
cloudlab:
  stitch-port:
      - name: AAA
        member-of:
          - UTAH
        device_name: dev1
      - name: BBB
        member-of:
          - UTAH
        device_name: dev2
  group:
    - name: UTAH
      producer-for:
        - fabric
    '''

    policy = load_policy(content=yaml_str)
    stitch_info = find_stitch_port(policy=policy,
                                   providers=['cloudlab', 'fabric'])

    stitch_port = stitch_info.stitch_port
    assert stitch_port['name'] == 'BBB'
    assert stitch_port.get('device_name') == 'dev2' or stitch_port['peer']['device_name'] == 'dev2'
    assert stitch_info.producer == 'cloudlab'
    assert stitch_info.consumer == 'fabric'

    policy = load_policy(content=yaml_str)
    stitch_info = find_stitch_port(policy=policy,
                                   providers=['cloudlab', 'fabric'], site='sitea')

    stitch_port = stitch_info.stitch_port
    assert stitch_port['name'] == 'AAA'
    assert stitch_port.get('device_name') == 'dev1' or stitch_port['peer']['device_name'] == 'dev1'
    assert stitch_info.producer == 'cloudlab'
    assert stitch_info.consumer == 'fabric'


def test_peering_and_duplicates():
    yaml_str = '''
fabric:
  stitch-port:
      - name: AAA
        member-of:
          - AWS
        device_name: dev1
      - name: AAA
        member-of:
            - AWS
        device_name: dev1
  group:
      - name: AWS
        producer-for:
          - sense
sense:
  stitch-port:
      - name: AAA
        member-of:
          - AWS
        region:
        device_name: dev3
        local_name: TenGigE0/0/0/11/3
  group:
      - name: AWS 
        profile: FABRIC-AWS-DX-VGW
        consumer-for:
          - fabric
        producer-for:
          - fabric
    '''

    policy = load_policy(content=yaml_str)
    stitch_infos = find_stitch_port_for_providers(policy, ['sense', 'fabric'])
    assert len(stitch_infos) == 2
    stitch_infos = peer_stitch_ports(stitch_infos)
    assert len(stitch_infos) == 1


def load_local_policy_using(providers):
    from amscrot.policy.policy_helper import load_policy

    policy = load_policy()

    from amscrot.policy.policy_helper import find_stitch_port_for_providers, peer_stitch_ports

    stitch_infos = find_stitch_port_for_providers(policy, providers)
    stitch_infos = peer_stitch_ports(stitch_infos)
    attrs = ["preference", "member-of", 'name']
    names = []

    for stitch_info in stitch_infos:
        names.append(stitch_info.stitch_port.get("name"))

        for attr in attrs:
            stitch_info.stitch_port.pop(attr, None)

        peer = stitch_info.stitch_port.get('peer', {})

        for attr in attrs:
            peer.pop(attr, None)

    import yaml
    from amscrot.util.constants import Constants

    for i, stitch_info in enumerate(stitch_infos):
        producer = stitch_info.producer
        consumer = stitch_info.consumer
        stitch_port = stitch_info.stitch_port
        c = {"config": [{Constants.NETWORK_STITCH_CONFIG: [{f"si_from_{names[i]}": {"producer": producer,
                                                                                    "consumer": consumer,
                                                                                    "stitch_port": stitch_port}}]}]}
        rep = yaml.dump(c, default_flow_style=False, sort_keys=False)
        rep = rep.replace('\n', '\n  ')
        print(rep)
    return stitch_infos


def test_load_local_policy_using_sense():
    providers = ['sense', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 2 
    providers = ['fabric', 'sense']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 2 


def test_load_local_policy_using_chi():
    providers = ['chi', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 3
    providers = ['fabric', 'chi']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 3


def test_load_local_policy_using_clab():
    providers = ['cloudlab', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 2
    providers = ['cloudlab', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 2


def test_load_local_policy_using_aws():
    providers = ['aws', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 1 
    providers = ['fabric', 'aws']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 1 


def test_load_local_policy_using_gcp():
    providers = ['gcp', 'fabric']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 1
    providers = ['fabric', 'gcp']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 1


def test_load_local_policy_zero():
    providers = ['gcp', 'aws']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 0
    providers = ['sense', 'chi']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 0
    providers = ['cloudlab', 'sense']
    stitch_infos = load_local_policy_using(providers)
    assert len(stitch_infos) == 0



from typing import List, Optional, Dict, Any, Union
from pydantic import BaseModel, Field


# -- Common base -------------------------------------------------------------

class ResourceBase(BaseModel):
    """Common fields shared by all first-class resource metadata objects."""
    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None


# -- Resource Models ----------------------------------------------------------

class Compute(ResourceBase):
    # Common / Shared
    architecture: Optional[str] = None
    cores: Optional[int] = None
    memory: Optional[str] = None
    
    # HPC Specific
    partition: Optional[str] = None
    nodes: Optional[int] = None
    tasks_per_node: Optional[int] = None
    gpus_per_node: Optional[int] = None
    gpu_type: Optional[str] = None
    memory_per_node: Optional[str] = None
    cpu_arch: Optional[str] = None

    # K8s Specific
    resources: Optional[Dict[str, Dict[str, str]]] = None # requests/limits
    replicas: Optional[int] = None
    node_selector: Optional[Dict[str, str]] = None
    container_runtime: Optional[str] = None


class Storage(ResourceBase):
    # Common
    type: Optional[str] = None # lustre, nfs, ebs
    mount_point: Optional[str] = None
    quota: Optional[str] = None
    
    # HPC Specific
    performance_tier: Optional[str] = None
    
    # K8s Specific
    kind: Optional[str] = None # PersistentVolumeClaim
    storage_class: Optional[str] = None
    size: Optional[str] = None
    access_modes: Optional[List[str]] = None


class BGPPeer(BaseModel):
    asn: int
    peer_ip: str
    local_ip: Optional[str] = None
    md5_auth: Optional[str] = None


class Layer2(BaseModel):
    vlan_ranges: Optional[List[str]] = None
    vlan_id: Optional[int] = None
    mtu: Optional[int] = None


class Layer3(BaseModel):
    ipv4_subnets: Optional[List[str]] = None
    ipv6_subnets: Optional[List[str]] = None
    bgp_peers: Optional[List[BGPPeer]] = None
    gateway: Optional[str] = None


class Network(ResourceBase):
    # HPC
    fabric: Optional[str] = None
    rdma_enabled: Optional[bool] = None
    external_connectivity: Optional[str] = None
    
    # K8s
    service: Optional[Dict[str, Any]] = None
    ingress: Optional[Dict[str, Any]] = None
    
    # Edge
    interface: Optional[str] = None
    bandwidth_limit: Optional[str] = None
    latency_tolerance: Optional[str] = None
    offline_support: Optional[bool] = None

    # Advanced L2/L3
    layer2: Optional[Layer2] = None
    layer3: Optional[Layer3] = None


class Allocation(ResourceBase):
    account: Optional[str] = None
    qos: Optional[str] = None
    walltime_limit: Optional[str] = None
    exclusive: Optional[bool] = None


class Data(ResourceBase):
    caching_policy: Optional[str] = None
    sync_interval: Optional[str] = None
    retention_policy: Optional[str] = None


class Operation(ResourceBase):
    strategy: Optional[str] = None
    max_unavailable: Optional[str] = None
    max_surge: Optional[str] = None


class Facility(ResourceBase):
    name: str  # required -- overrides the optional base field
    compute: Optional[List[Compute]] = None
    storage: Optional[List[Storage]] = None
    networks: Optional[List[Network]] = None
    allocations: Optional[List[Allocation]] = None
    operations: Optional[List[Operation]] = None
    data: Optional[List[Data]] = None


# Service Models

class Service(BaseModel):
    name: str
    endpoint_uri: str
    status: str
    capabilities: Optional[List[str]] = None
    allocated: Optional[bool] = None


class Group(BaseModel):
    name: str


class Policy(BaseModel):
    profile: str
    group: str
    preference: int = Field(ge=0, le=1000)


class ServiceClient(BaseModel):
    name: str
    serviceRef: str # Reference to Service.name
    facilities: List[Facility]


class MetadataDump(BaseModel):
    groups: List[Group]
    policies: Dict[str, Policy]
    services: List[Service]
    service_clients: List[ServiceClient]



from amscrot.exceptions import ResourceTypeNotSupported, ProviderException
from amscrot.provider.api.provider import Provider
from amscrot.util.constants import Constants
from amscrot.util.utils import get_logger
from . import gcp_constants

logger = get_logger()


class GcpProvider(Provider):
    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        self.supported_resources = [Constants.RES_TYPE_NETWORK.lower()]

    @property
    def project(self):
        return self.config.get(gcp_constants.PROJECT)

    @property
    def service_key_path(self):
        return self.config.get(gcp_constants.SERVICE_KEY_PATH)

    def setup_environment(self):
        normalized_config = {}

        for k, v in self.config.items():
            normalized_config[k.upper()] = v

        self.config = normalized_config
        assert self.config.get(gcp_constants.SERVICE_KEY_PATH)
        assert self.config.get(gcp_constants.PROJECT)

        for attr in [gcp_constants.PROJECT, gcp_constants.SERVICE_KEY_PATH]:
            if self.config.get(attr) is None:
                raise ProviderException(f"{self.name}: Expecting a value for {attr}")

        skey = self.config[gcp_constants.SERVICE_KEY_PATH]

        from amscrot.util.utils import can_read, absolute_path

        skey = absolute_path(skey)

        if not can_read(skey):
            raise ProviderException(f"{self.name}: unable to read service key in {skey}")

        try:
            with open(skey, 'r') as fp:
                import json

                json.load(fp)
        except Exception as e:
            raise ProviderException(f"{self.name}:Unable to parse {skey} to json:{e}")

        self.config[gcp_constants.SERVICE_KEY_PATH] = skey

    def do_add_resource(self, *, resource: dict):
        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)

        if rtype not in self.supported_resources:
            raise ResourceTypeNotSupported(f"{rtype} for {label}")

        name_prefix = resource[Constants.RES_NAME_PREFIX]
        name_prefix = name_prefix.replace("_", "-")
        net_name = f'{self.name}-{name_prefix}'

        layer3 = resource.get(Constants.RES_LAYER3)
        peering = resource.get(Constants.RES_PEERING)

        from .gcp_network import GcpNetwork
        from amscrot.policy.policy_helper import get_stitch_port_for_provider

        stitch_port = get_stitch_port_for_provider(resource=resource, provider=self.type)
        net = GcpNetwork(label=label, name=net_name, provider=self, layer3=layer3, peering=peering, stitch_port=stitch_port)
        self._networks.append(net)

        if self.resource_listener:
            self.resource_listener.on_added(source=self, provider=self, resource=net)

    def do_create_resource(self, *, resource: dict):
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)

        for net in [net for net in self._networks if net.label == label]:
            self.logger.debug(f"Creating network: {vars(net)}")
            net.create()
            self.logger.debug(f"Created network: {vars(net)}")

            if self.resource_listener:
                self.resource_listener.on_created(source=self, provider=self, resource=net)

    def do_delete_resource(self, *, resource: dict):
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)

        if rtype == Constants.RES_TYPE_NODE.lower():
            # DO NOTHING
            return

        name_prefix = resource[Constants.RES_NAME_PREFIX]
        name_prefix = name_prefix.replace("_", "-")
        net_name = f'{self.name}-{name_prefix}'
        logger.debug(f"Deleting network: {net_name}")

        from .gcp_network import GcpNetwork
        from amscrot.policy.policy_helper import get_stitch_port_for_provider

        layer3 = resource.get(Constants.RES_LAYER3)
        peering = resource.get(Constants.RES_PEERING)
        stitch_port = get_stitch_port_for_provider(resource=resource, provider=self.type)
        net = GcpNetwork(label=label, name=net_name, provider=self, layer3=layer3, peering=peering, stitch_port=stitch_port)
        net.delete()
        logger.debug(f"Done Deleting network: {net_name}")

        if self.resource_listener:
            self.resource_listener.on_deleted(source=self, provider=self, resource=net)



from .parser import Parser
from .config_models import ResourceConfig, ProviderConfig
from typing import List, Union, Dict


class WorkflowConfig:
    def __init__(self, *, provider_configs: List[ProviderConfig], resource_configs: List[ResourceConfig]):
        self.provider_configs = provider_configs
        self.resource_configs = resource_configs

    def get_provider_configs(self) -> List[ProviderConfig]:
        return self.provider_configs

    def get_resource_configs(self) -> List[ResourceConfig]:
        return self.resource_configs

    @staticmethod
    def parse(*, dir_path: Union[str, None] = None, content: Union[str, None] = None,
              var_dict: Union[Dict, None] = None):
        provider_configs, resource_configs = Parser.parse(dir_path=dir_path, content=content, var_dict=var_dict)
        return WorkflowConfig(provider_configs=provider_configs, resource_configs=resource_configs)



import time

import boto3

from amscrot.util.utils import get_logger
from amscrot.util.constants import Constants
from .aws_constants import *
from .aws_exceptions import AwsException

logger = get_logger()


def create_ec2_client(*, region: str, access_key: str, secret_key: str):
    ec2_client = boto3.client(
        'ec2',
        region_name=region,
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key
    )

    return ec2_client


def is_vpc_available(*, ec2_client, vpc_id: str):
    response = ec2_client.describe_vpcs(VpcIds=[vpc_id])

    if response and isinstance(response, dict) and 'Vpcs' in response:
        vpc = next(iter(response['Vpcs']), None)

        if vpc:
            state = vpc['State']

            if state == 'available':
                return True

            logger.warning(f'found vpc {vpc_id} with state={state}')
        else:
            raise AwsException(f"Vpc not found {vpc_id}")

    return False


def find_route_tables(*, ec2_client, vpc_id):
    response = ec2_client.describe_route_tables()
    route_tables = []

    if response and isinstance(response, dict) and 'RouteTables' in response:
        for route_table in response['RouteTables']:
            if route_table['VpcId'] == vpc_id:
                route_tables.append(route_table)

    return route_tables


def create_subnet_if_needed(*, ec2_client, cidr, vpc_id):
    response = ec2_client.describe_subnets()
    subnet = next(
        filter(lambda sub: sub['CidrBlock'] == cidr and sub['VpcId'] == vpc_id, response['Subnets']),
        None
    )

    if not subnet:
        logger.info(f'Creating subnet {cidr}')

        response = ec2_client.create_subnet(
            CidrBlock=cidr,
            VpcId=vpc_id
        )

        subnet = response['Subnet']

    logger.info(f'Got Subnet {subnet}')

    state = subnet['State']
    subnet_id = subnet['SubnetId']

    if state == 'available':
        return subnet_id

    for i in range(RETRY):
        response = ec2_client.describe_subnets(SubnetIds=[subnet_id])
        subnet = next(filter(lambda sub: sub['CidrBlock'] == cidr and sub['VpcId'] == vpc_id, response['Subnets']))
        state = subnet['State']
        logger.info(f'Waiting on subnet {subnet_id}: state={state}')

        if state == 'available':
            return subnet_id

        time.sleep(20)

    raise AwsException(f'Timed out. subnet {subnet_id}:state={state}')


# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/enable_vgw_route_propagation.html
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/describe_route_tables.html

def print_route_tables(ec2_client, vpc_id):
    route_tables = find_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)
    print()
    print("Number of route tables", len(route_tables))

    for route in route_tables:
        print()
        print("Begin Routetable")
        print(route)

        print("PropagatingVgws", route['PropagatingVgws'])

        for asso in route['Associations']:
            print('ASSOCIATION=', asso, "::::")
        print("END Routetable")

    print()


def create_route_table_if_needed(*, ec2_client, cidr, vpc_id, vpn_id):
    subnet_id = create_subnet_if_needed(ec2_client=ec2_client, cidr=cidr, vpc_id=vpc_id)
    route_tables = find_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)

    for route_table in route_tables:
        if {'GatewayId': vpn_id} in route_table['PropagatingVgws']:
            logger.info(f'Already associated to vpn {vpn_id}')
            print_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)
            return route_table

    for route_table in route_tables:
        for asso in route_table['Associations']:
            if asso.get('SubnetId') == subnet_id:
                response = ec2_client.disassociate_route_table(
                    AssociationId=asso['RouteTableAssociationId']
                )

                logger.info(f'disassociated_route_table from subnet={subnet_id}:response={response}')
                break

    logger.info(f'Creating route_table vpc={vpc_id}')
    response = ec2_client.create_route_table(VpcId=vpc_id)
    route_table = response['RouteTable']
    route_table_id = route_table['RouteTableId']
    response = ec2_client.associate_route_table(RouteTableId=route_table_id, SubnetId=subnet_id)
    logger.info(f'associate_created_route_table:response={response}')
    response = ec2_client.enable_vgw_route_propagation(GatewayId=vpn_id, RouteTableId=route_table_id)
    logger.info(f'enable_vgw_route_propagation_to_created_table:response={response}')
    print_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)
    return route_table


def delete_route_table_if_needed(*, ec2_client, vpc_id, route_table_id):
    logger.info(f'deleting route_table:{route_table_id}')
    route_tables = find_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)
    route_table = next(filter(lambda rt: rt['RouteTableId'] == route_table_id, route_tables), None)

    if not route_table:
        logger.info(f'route_table not found:{route_table_id}')
        return

    for route_table in route_tables:
        for asso in route_table['Associations']:
            if asso.get('SubnetId'):
                subnet_id = asso.get('SubnetId')
                response = ec2_client.disassociate_route_table(
                    AssociationId=asso['RouteTableAssociationId']
                )

                logger.info(f'disassociated_route_table from subnet={subnet_id}: response={response}')

    response = ec2_client.delete_route_table(RouteTableId=route_table_id)
    logger.info(f'deleted route_table:{route_table_id}:response={response}')

    for i in range(RETRY):
        route_tables = find_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)
        route_table = next(filter(lambda rt: rt['RouteTableId'] == route_table_id, route_tables), None)

        if not route_table:
            break
        logger.info(f'waiting on deleting route_table:{route_table_id}:response={response}')
        time.sleep(20)

    logger.info(f'done deleting route_table:{route_table_id}:response={response}')
    print_route_tables(ec2_client=ec2_client, vpc_id=vpc_id)


def find_available_dx_connection(*, direct_connect_client, name: str):
    response = direct_connect_client.describe_connections()

    if not isinstance(response, dict) and 'connections' not in response:
        raise AwsException(f'did not find dx connection {name}')

    # connection = next(filter(lambda con: con['connectionName'] == name, response['connections']), None)
    connections = list(filter(lambda con: con['connectionName'] == name, response['connections']))

    if not connections:
        raise AwsException(f'did not find dx connection {name}')

    for connection in connections:
        logger.info(f'Found dx connection {connection}')
    
        state = connection['connectionState']
        vlan = connection['vlan']
        connection_id = connection['connectionId']
        
        if state in ['down', 'deleting', 'deleted', 'rejected', 'unknown']:
            continue
    
        if state == 'available':
            return connection_id, vlan
    
        if state == 'ordering':
            response = direct_connect_client.confirm_connection(connectionId=connection_id)
            logger.info(f'response from confirm dx connection {response}')
    
        for i in range(RETRY):
            response = direct_connect_client.describe_connections(connectionId=connection_id)
            connection = next(filter(lambda con: con['connectionName'] == name, response['connections']))
            state = connection['connectionState']
            logger.info(f'state={state}. dx connection {connection}')
    
            if state == 'available':
                return connection_id, vlan
    
            time.sleep(20)

    raise AwsException(f'Timed out. dx connection {name}:state={state}')


def find_vpn_gateway(*, ec2_client, name: str):
    response = ec2_client.describe_vpn_gateways()

    if isinstance(response, dict) and 'VpnGateways' in response:
        for gw in response['VpnGateways']:
            if 'Tags' in gw and gw['Tags']:
                name_tag = next(filter(lambda t: t['Key'] == 'Name' and t['Value'] == name, gw['Tags']), None)

                if name_tag:
                    return gw['VpnGatewayId']

    return None


def _find_vpn_gateway_by_id(*, ec2_client, vpn_id: str):
    response = ec2_client.describe_vpn_gateways()

    if response and isinstance(response, dict) and 'VpnGateways' in response:
        for gw in response['VpnGateways']:
            if gw['VpnGatewayId'] == vpn_id:
                return gw

    return None


def attach_vpn_gateway_if_needed(*, ec2_client, vpn_id: str, vpc_id: str):
    vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)
    attachments = vpn_gateway['VpcAttachments']
    attachment = next(filter(lambda at: at['VpcId'] == vpc_id, attachments), None)

    if not attachment or attachment['State'] in ['detaching', 'detached']:
        response = ec2_client.attach_vpn_gateway(
            VpcId=vpc_id,
            VpnGatewayId=vpn_id
        )

        attachment = response['VpcAttachment']

    state = attachment['State']

    if state == 'attached':
        logger.info(f"VPN {vpn_id}:state={state}")
        return

    for i in range(RETRY):
        vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)
        attachments = vpn_gateway['VpcAttachments']
        attachment = next(filter(lambda at: at['VpcId'] == vpc_id, attachments))
        state = attachment['State']

        if state == 'attached':
            logger.info(f"VPN {vpn_id}:state={state}")
            return

        logger.info(f"Waiting on attaching VPN {vpn_id}:state={state}")
        time.sleep(20)

    raise AwsException(f"Timed out on attaching vpn_gateway: state={state}")


def detach_vpn_gateway_if_needed(*, ec2_client, vpn_id: str, vpc_id: str):
    vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)
    attachments = vpn_gateway['VpcAttachments']
    attachment = None

    for a in attachments:
        if a['VpcId'] == vpc_id:
            attachment = a

    if not attachment:
        return

    try:
        logger.info(f"Detaching vpn:id={vpn_id}")
        ec2_client.detach_vpn_gateway(
            VpcId=vpc_id,
            VpnGatewayId=vpn_id
        )
    except Exception as e:
        logger.warning(f"failed to detach vpn: {e}")

    state = None

    for i in range(RETRY):
        vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)
        attachments = vpn_gateway['VpcAttachments']
        state = None

        for attachment in attachments:
            if attachment['VpcId'] == vpc_id:
                state = attachment['State']
                break

        if not state or state == 'detached':
            return

        logger.info(f"Waiting on detached vpn state={state}")
        time.sleep(20)

    raise AwsException(f"Timed out on detaching vpn_gateway: state={state}")


def create_vpn_gateway(*, ec2_client, name: str, amazon_asn: int):
    response = ec2_client.create_vpn_gateway(
        Type='ipsec.1',
        TagSpecifications=[
            {
                'ResourceType': 'vpn-gateway',
                'Tags': [
                    {
                        'Key': 'Name',
                        'Value': name
                    },
                ]
            },
        ],
        AmazonSideAsn=amazon_asn
    )

    vpn_gateway = response['VpnGateway']
    vpn_id = vpn_gateway['VpnGatewayId']
    state = vpn_gateway['State']

    if state == 'available':
        logger.info(f"Returning VPN {name}:state={state}")
        return vpn_id

    for i in range(RETRY):
        vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)

        if vpn_gateway:
            state = vpn_gateway['State']

            if state == 'available':
                return vpn_id

        logger.info(f"Waiting on VPN {name}:state={state}")
        time.sleep(20)

    raise AwsException(f"Timed out on creating vpn_gateway: state={state}")


def delete_vpn_gateway(*, ec2_client, name: str):
    vpn_id = find_vpn_gateway(ec2_client=ec2_client, name=name)

    if not vpn_id:
        return

    ec2_client.delete_vpn_gateway(VpnGatewayId=vpn_id)
    state = None

    for i in range(RETRY):
        vpn_gateway = _find_vpn_gateway_by_id(ec2_client=ec2_client, vpn_id=vpn_id)

        if vpn_gateway:
            state = vpn_gateway['State']

        if not state or state == 'deleted':
            return vpn_id

        time.sleep(20)

    raise AwsException(f"Timed out on creating vpn_gateway: state={state}")


def create_direct_connect_client(*, region: str, access_key: str, secret_key: str):
    direct_connect_client = boto3.client(
        'directconnect',
        region_name=region,
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key
    )

    return direct_connect_client


def find_direct_connect_gateway_by_name(*, direct_connect_client, gateway_name: str):
    response = direct_connect_client.describe_direct_connect_gateways()

    if response and isinstance(response, dict) and 'directConnectGateways' in response:
        for gw in response['directConnectGateways']:
            if gw['directConnectGatewayName'] == gateway_name:
                return gw['directConnectGatewayId']

    return None


def create_direct_connect_gateway(*, direct_connect_client, gateway_name: str, amazon_asn: int):
    direct_connect_gateway_id = find_direct_connect_gateway_by_name(
        direct_connect_client=direct_connect_client,
        gateway_name=gateway_name)

    if not direct_connect_gateway_id:
        response = direct_connect_client.create_direct_connect_gateway(
            directConnectGatewayName=gateway_name,
            amazonSideAsn=amazon_asn
        )

        logger.info(f"Created directConnectGateway {response}")
        direct_connect_gateway_id = response['directConnectGateway']['directConnectGatewayId']

    return direct_connect_gateway_id


def delete_direct_connect_gateway(*, direct_connect_client, gateway_name: str):
    direct_connect_gateway_id = find_direct_connect_gateway_by_name(
        direct_connect_client=direct_connect_client,
        gateway_name=gateway_name)

    if direct_connect_gateway_id:
        response = direct_connect_client.delete_direct_connect_gateway(
            directConnectGatewayId=direct_connect_gateway_id
        )

        logger.info(f"Deleted directConnectGateway {response}")

    return direct_connect_gateway_id


# 'virtualInterfaceState':
# 'confirming'|'verifying'|'pending'|'available'|'down'|'deleting'|'deleted'|'rejected'|'unknown',
def create_private_virtual_interface(*,
                                     direct_connect_client,
                                     vpn_gateway_id: str,
                                     connection_id,
                                     vlan,
                                     peering,
                                     vif_name: str):
    response = direct_connect_client.describe_virtual_interfaces()
    details = {}

    if isinstance(response, dict) and 'virtualInterfaces' in response:
        for vif in response['virtualInterfaces']:
            if vif['virtualGatewayId'] == vpn_gateway_id and vif['virtualInterfaceName'] == vif_name:
                logger.info(f"Found existing private virtual interface {vif_name}")

                for k in VIF_DETAILS:
                    details[k] = vif[k]

                break

    if not details:
        logger.info(f"Creating private virtual interface {vif_name}:connection_id={connection_id}:vlan={vlan}")

        bgp_asn = peering.attributes.get(Constants.RES_LOCAL_ASN)
        local_address = peering.attributes.get(Constants.RES_LOCAL_ADDRESS)
        remote_address = peering.attributes.get(Constants.RES_REMOTE_ADDRESS)
        bgp_key = peering.attributes.get(Constants.RES_SECURITY)

        vif = direct_connect_client.create_private_virtual_interface(
            connectionId=connection_id,
            newPrivateVirtualInterface={
                'virtualInterfaceName': vif_name,
                'vlan': vlan,
                'asn': bgp_asn,  # remote oess asn
                'mtu': 9001,
                'authKey': bgp_key,
                'amazonAddress': remote_address,   # local_address,
                'customerAddress': local_address,  # remote_address,
                'addressFamily': 'ipv4',
                'virtualGatewayId': vpn_gateway_id,
                'enableSiteLink': False
            }
        )

        for k in VIF_DETAILS:
            details[k] = vif[k]

    if details[VIF_STATE] == 'available':
        logger.info(f"Private virtual interface {vif_name} is {details[VIF_STATE]}")
        return details

    for i in range(RETRY):
        logger.warning(f"Waiting on private virtual interface {vif_name}:state={details[VIF_STATE]}:attempt={i + 1}")
        time.sleep(20)

        response = direct_connect_client.describe_virtual_interfaces()
        vif = next(filter(lambda v: v[VIF_ID] == details[VIF_ID], response['virtualInterfaces']))

        for k in VIF_DETAILS:
            details[k] = vif[k]

        state = details[VIF_STATE]

        if state == 'available':
            break

    if details[VIF_STATE] != 'available':
        raise AwsException(f"Virtual interface {vif_name}:state={details[VIF_STATE]}")

    logger.info(f"Private virtual interface {vif_name} is {details[VIF_STATE]}")
    return details


def delete_private_virtual_interface(*,
                                     direct_connect_client,
                                     vif_name: str):
    response = direct_connect_client.describe_virtual_interfaces()
    details = {}

    if isinstance(response, dict) and 'virtualInterfaces' in response:
        for vif in response['virtualInterfaces']:
            if vif['virtualInterfaceName'] == vif_name:
                logger.info(f"Found existing private virtual interface {vif_name}")

                for k in VIF_DETAILS:
                    details[k] = vif[k]

                break

    if not details:
        return

    if details[VIF_STATE] == 'deleted':
        logger.info(f"Private virtual interface {vif_name} is {details[VIF_STATE]}")
        return details

    logger.info(f"Deleting private virtual interface {vif_name}:{details}")
    direct_connect_client.delete_virtual_interface(
        virtualInterfaceId=details[VIF_ID]
    )

    for i in range(RETRY):
        logger.warning(f"Waiting on private virtual interface {vif_name}:state={details[VIF_STATE]}:attempt={i + 1}")
        time.sleep(20)

        response = direct_connect_client.describe_virtual_interfaces()
        vif = next(filter(lambda v: v[VIF_ID] == details[VIF_ID], response['virtualInterfaces']))

        for k in VIF_DETAILS:
            details[k] = vif[k]

        state = details[VIF_STATE]

        if state == 'deleted':
            break

    if details[VIF_STATE] != 'deleted':
        raise AwsException(f"Virtual interface {vif_name}:state={details[VIF_STATE]}")

    logger.info(f"Private virtual interface {vif_name} is {details[VIF_STATE]}")
    return details


def find_association_dxgw_vpn(*, direct_connect_client, direct_connect_gateway_id: str, vpn_id: str):
    response = direct_connect_client.describe_direct_connect_gateway_associations(
        virtualGatewayId=vpn_id,
        directConnectGatewayId=direct_connect_gateway_id,
    )

    if response and isinstance(response, dict) and 'directConnectGatewayAssociations' in response:
        associations = response['directConnectGatewayAssociations']

        if associations:
            return associations[0]

    return None


def find_association_dxgw_vpn_id(*, direct_connect_client, direct_connect_gateway_id: str, vpn_id: str):
    association = find_association_dxgw_vpn(
        direct_connect_client=direct_connect_client,
        direct_connect_gateway_id=direct_connect_gateway_id,
        vpn_id=vpn_id
    )

    if association:
        return association['associationId']

    return None


def associate_dxgw_vpn(direct_connect_client, direct_connect_gateway_id, vpn_id: str):
    association = find_association_dxgw_vpn(
        direct_connect_client=direct_connect_client,
        direct_connect_gateway_id=direct_connect_gateway_id,
        vpn_id=vpn_id
    )

    if association and association['associationState'] == 'associated':
        logger.info(f'association is associated:{association}')
        return association['associationId']

    if not association:
        response = direct_connect_client.create_direct_connect_gateway_association(
            directConnectGatewayId=direct_connect_gateway_id,
            virtualGatewayId=vpn_id
        )
        association = response['directConnectGatewayAssociation']

    state = association['associationState']

    if state == 'associated':
        logger.info(f'association is associated:{association}')
        return association['associationId']

    for i in range(RETRY):
        logger.warning(f'Waiting on association. state={state}: association:{association}')
        time.sleep(20)

        association = find_association_dxgw_vpn(
            direct_connect_client=direct_connect_client,
            direct_connect_gateway_id=direct_connect_gateway_id,
            vpn_id=vpn_id
        )

        state = association['associationState']

        if state == 'associated':
            logger.info(f'association is associated:{association}')
            return association['associationId']

    raise AwsException(f"Timed out on creating direct_connect_gateway_association: state={state}")


def dissociate_dxgw_vpn(*, direct_connect_client, association_id: str):
    response = direct_connect_client.delete_direct_connect_gateway_association(
        associationId=association_id
    )

    association = response['directConnectGatewayAssociation']

    if not association or not isinstance(association, dict) or 'associationState' not in association:
        return

    state = association['associationState']

    if state == 'disassociated':
        return

    direct_connect_gateway_id = association['directConnectGatewayId']
    vpn_id = association['virtualGatewayId']

    for i in range(RETRY):
        association = find_association_dxgw_vpn(
            direct_connect_client=direct_connect_client,
            direct_connect_gateway_id=direct_connect_gateway_id,
            vpn_id=vpn_id
        )

        if not association or not isinstance(association, dict) or 'associationState' not in association:
            return

        logger.info(f'checking if association state is disassociated: {association}')
        state = association['associationState']

        if state == 'disassociated':
            return

        time.sleep(20)

    raise AwsException(f"Timed out on deleting direct_connect_gateway_association:id={association_id}:state={state}")



from amscrot.util.constants import Constants
import enum

AUTH_ENDPOINT = "auth_endpoint"
API_ENDPOINT = "api_endpoint"
CLIENT_ID = "client_id"
USERNAME = "username"
PASSWORD = "password"
SECRET = "secret"
SENSE_SLICE_PRIVATE_KEY_LOCATION = "slice-private-key-location"

SENSE_CONF_ATTRS = [AUTH_ENDPOINT, API_ENDPOINT, CLIENT_ID, USERNAME, PASSWORD, SECRET]

SENSE_PROFILE_UID = "service_profile_uuid"
SENSE_ALIAS = "alias"
SENSE_EDIT = "options"
SENSE_URI = 'uri'
# SENSE_ID = 'id'
SENSE_PATH = 'path'
SENSE_VLAN_TAG = 'vlan_tag'

SENSE_DTN = 'With Host'
SENSE_DTN_IP = 'IP Address'

SERVICE_INSTANCE_KEYS = ['intents', 'alias', 'referenceUUID', 'state', 'owner', 'lastState', 'timestamp', 'archived']

SENSE_CUSTOMER_ASN = "customer_asn"
SENSE_AMAZON_ASN = "amazon_asn"
SENSE_GOOGLE_ASN = "google_asn"
SENSE_CUSTOMER_IP = "customer_ip"
SENSE_AMAZON_IP = "amazon_ip"
SENSE_GOOGLE_IP = "google_ip"
SENSE_AUTHKEY = "authkey"
SENSE_TO_HOSTED_CONN = "to_hosted_conn"

SENSE_AWS_PEERING_MAPPING = {
    Constants.RES_LOCAL_ASN: SENSE_CUSTOMER_ASN,
    Constants.RES_LOCAL_ADDRESS: SENSE_CUSTOMER_IP,

    Constants.RES_REMOTE_ASN: SENSE_AMAZON_ASN,
    Constants.RES_REMOTE_ADDRESS: SENSE_AMAZON_IP,

    Constants.RES_SECURITY: SENSE_AUTHKEY,

    Constants.RES_ID: SENSE_TO_HOSTED_CONN,
}

SENSE_GCP_PEERING_MAPPING = {
    Constants.RES_LOCAL_ASN: SENSE_CUSTOMER_ASN,
    Constants.RES_LOCAL_ADDRESS: SENSE_CUSTOMER_IP,

    Constants.RES_REMOTE_ASN: SENSE_GOOGLE_ASN,
    Constants.RES_REMOTE_ADDRESS: SENSE_GOOGLE_IP,

    Constants.RES_SECURITY: SENSE_AUTHKEY
}

# This is what we get from node_details from manifest for AWS ...
# 'Public IP': '18.215.246.8', 'Node Name': 'VM-1',
# 'Key Pair': 'keypair+kp-sense', 'Image': 'image+ami-052efd3df9dad4825'}

# ubuntu/images/hvm-ssd/ubuntu-jammy-22.04-amd64-server-20220609
# ami-052efd3df9dad4825
# ami-052efd3df9dad4825
# Canonical, Ubuntu, 22.04 LTS, amd64 jammy image build on 2022-06-09

SENSE_KEYPAIR = 'Key Pair'
SENSE_PUBLIC_IP = 'Public IP'
SENSE_PRIVATE_IP = 'Private IP'
SENSE_NODE_NAME = 'Node Name'
SENSE_IMAGE = 'Image'

SENSE_RETRY = 50

class SupportedCloud(str, enum.Enum):
    """
    The Cloud supported by SENSE.
    """
    GCP = "GOOGLE"
    AWS = "AMAZON"
    DTN = "DTNL2"



import json
import uuid
from types import SimpleNamespace

from sense.client.discover_api import DiscoverApi
from sense.client.profile_api import ProfileApi
from sense.client.workflow_combined_api import WorkflowCombinedApi

from .sense_client import get_client

from .sense_constants import *
from .sense_exceptions import SenseException

from amscrot.util.utils import get_logger

logger = get_logger()


def get_image_info(image_spec, attr=None):
    import os

    path_file = os.path.join(os.path.dirname(__file__), 'inventory', 'sense_image_info.json')

    with open(path_file, 'r') as fp:
        infos = json.load(fp)

    if not attr:
        return infos.get(image_spec)

    return infos.get(image_spec, dict()).get(attr)


def describe_profile(*, client=None, uuid: str):
    client = client or get_client()
    profile_api = ProfileApi(req_wrapper=client)
    profile_details = profile_api.profile_describe(uuid)
    #profile_details = json.loads(profile_details, object_hook=lambda dct: SimpleNamespace(**dct))

    if hasattr(profile_details, "edit"):
        from requests import utils

        for ns in profile_details.edit:
            ns.path = utils.unquote(ns.path)
            ns.valid = utils.unquote(ns.valid)

    return profile_details


def populate_options_using_interfaces(options, interfaces, edit_uri_entries):
    if len(interfaces) > len(edit_uri_entries):
        raise SenseException("invalid number of interfaces")

    index_used = False

    for interface in interfaces:
        if interface.get(Constants.RES_INDEX):
            index_used = True
            break

    if not index_used:
        for idx, interface in enumerate(interfaces):
            interface[Constants.RES_INDEX] = idx
    else:
        for interface in interfaces:
            if not interface.get(Constants.RES_INDEX):
                raise SenseException(f"index missing {interface}")

    for interface in interfaces:
        idx = interface.get(Constants.RES_INDEX)

        if idx < 0 or idx >= len(edit_uri_entries):
            raise SenseException("bad interface index")

        name = interface.get(Constants.RES_ID)
        path_prefix = f"data.connections[0].terminals[{idx}]."

        if name:
            path = path_prefix + SENSE_URI
            options.append({path: name})

        vlan_range = interface.get("vlan_range", None)

        if vlan_range:
            vlan_path = path_prefix + SENSE_VLAN_TAG
            vlan_range_str = str(vlan_range).replace(" ", "").replace(",", "-")
            options.append({vlan_path: vlan_range_str[1:-1]})


def get_profile_uuid(*, client=None, profile):
    try:
        uuid.UUID(profile)
        return profile
    except Exception:
        pass
    client = client or get_client()
    profile_api = ProfileApi(req_wrapper=client)

    try:
        profile = profile_api.profile_search_get_with_http_info(search=profile)
        #profile = json.loads(profile, object_hook=lambda dct: SimpleNamespace(**dct))
        profile_uuid = profile.get('uuid')
        return profile_uuid
    except Exception as e:
        raise SenseException(f"Exception searching for profile:{e}")


def create_instance(*, client=None, bandwidth, profile, alias, layer3, peering, interfaces):
    client = client or get_client()
    profile_uuid = get_profile_uuid(client=client, profile=profile)
    workflow_api = WorkflowCombinedApi(req_wrapper=client)
    workflow_api.instance_new()
    intent = {SENSE_PROFILE_UID: profile_uuid, "alias": alias}
    edit_entries = []
    profile_details = describe_profile(client=client, uuid=profile_uuid)

    logger.debug(f'Profile Details: {profile_details}')

    if hasattr(profile_details, "edit"):
        edit_entries = profile_details.edit
        temp_entries = [e.__dict__ for e in edit_entries]
        logger.info(f'Edit Entries: {json.dumps(temp_entries, indent=2)}')

    edit_uri_entries = [e for e in edit_entries if e.path.endswith(SENSE_URI)]
    options = []

    if interfaces:
        populate_options_using_interfaces(options, interfaces, edit_uri_entries)

    if bandwidth:
        options.append({f"data.connections[{0}].bandwidth.capacity": bandwidth})

    if peering:
        try:
            gateway_type = profile_details.intent.data.gateways[0].type.upper()
            if "GCP" in gateway_type:
                peering_mapping = SENSE_GCP_PEERING_MAPPING
            elif "AWS" in gateway_type:
                peering_mapping = SENSE_AWS_PEERING_MAPPING
            else:
                raise SenseException(f"Was not able to figure out peering mapping for {gateway_type}")
        except Exception as e:
            raise SenseException(f"Was not able to figure out peering mapping:{str(e)}")

        for k, v in peering_mapping.items():
            if k in peering.attributes:
                path = "data.gateways[0].connects[0]." + v
                options.append({path: peering.attributes[k]})

    if layer3:
        edit_entry_paths = [e.path for e in edit_entries]
        subnet = layer3.attributes[Constants.RES_SUBNET]

        from ipaddress import IPv4Network

        subnet = IPv4Network(subnet)

        if "data.subnets[0].cidr" in edit_entry_paths:
            options.append({f"data.subnets[0].cidr": str(subnet)})
            vpc_subnet = subnet

            for i in [8, 4, 2]:
                if subnet.prefixlen - i > 0:
                    vpc_subnet = subnet.supernet(i)
                    break

            options.append({"data.cidr": str(vpc_subnet)})

    if options:
        query = dict([("ask", "edit"), ("options", options)])
        intent["queries"] = [query]

    logger.info(f'Intent: {json.dumps(intent, indent=2)}')
    intent = json.dumps(intent)

    import time
    from random import randint

    for attempt in range(SENSE_RETRY):
        try:
            logger.info(f"creating instance: {alias}:attempt={attempt + 1}")
            response = workflow_api.instance_create(intent)  # service_uuid, intent_uuid, queries, model
            #temp = json.loads(response)
            status = workflow_api.instance_get_status()
            return response['service_uuid'], status
        except Exception as e:
            logger.warning(f"exception while creating instance {e}")

        time.sleep(randint(30, 35))

    raise SenseException(f"could not create instance {alias}")


def instance_operate(*, client=None, si_uuid, action='provision'):
    import time
    from random import randint

    client = client or get_client()
    workflow_api = WorkflowCombinedApi(req_wrapper=client)
    status = workflow_api.instance_get_status(si_uuid=si_uuid)

    if "CREATE - COMMITTING" not in status or 'CANCEL - READY' == status:
        try:
            time.sleep(randint(5, 10))
            workflow_api.instance_operate(action, si_uuid=si_uuid, async_req=True)
        except Exception as e:
            logger.warning(f"exception from instance_operate {e}")


def wait_for_instance_operate(*, client=None, si_uuid,
                              statuses=('CREATE - READY', 'REINSTATE - READY')):
    import time
    from random import randint

    client = client or get_client()
    workflow_api = WorkflowCombinedApi(req_wrapper=client)

    for attempt in range(SENSE_RETRY):
        try:
            status = workflow_api.instance_get_status(si_uuid=si_uuid)
            logger.info(f"Waiting on {statuses}: status={status}:attempt={attempt} out of {SENSE_RETRY}")

            if status in statuses:
                return status

            if 'FAILED' in status:
                return status
        except Exception as e:
            logger.warning(f"exception from  instance_get_status {e}")
            pass

        logger.info(f"Waiting on {statuses}: going to sleep attempt={attempt}")
        time.sleep(randint(30, 35))

    return workflow_api.instance_get_status(si_uuid=si_uuid)


def delete_instance(*, client=None, si_uuid):
    import time
    import random

    client = client or get_client()
    workflow_api = WorkflowCombinedApi(req_wrapper=client)

    status = workflow_api.instance_get_status(si_uuid=si_uuid)

    if 'error' in status:
        raise SenseException("error deleting got " + status)

    if 'FAILED' in status:
        raise SenseException(f'cannot delete instance - contact admin. {status}')

    if "CREATE - COMPILED" in status:
        time.sleep(random.randint(5, 30))

        workflow_api.instance_delete(si_uuid=si_uuid)
        return

    if "CANCEL" not in status:
        if 'CREATE' not in status and 'REINSTATE' not in status and 'MODIFY' not in status:
            raise ValueError(f"cannot cancel an instance in '{status}' status...")

        time.sleep(random.randint(5, 30))

        if 'READY' not in status:
            workflow_api.instance_operate('cancel', si_uuid=si_uuid, sync='false', force='true')
        else:
            workflow_api.instance_operate('cancel', si_uuid=si_uuid, sync='false')

    for attempt in range(SENSE_RETRY):
        # This sleep is here to workaround issue where CANCEL-READY shows up prematurely.
        time.sleep(random.randint(30, 35))

        status = workflow_api.instance_get_status(si_uuid=si_uuid)
        # print("LOOPING:DELETE:Status=", status, "attempt=", attempt)
        logger.info(f"Waiting on CANCEL-READY: status={status}:attempt={attempt} out of {SENSE_RETRY}")

        if 'CANCEL - READY' in status:  # This got triggered very quickly ...
            break

        if 'FAILED' in status:
            break

    if 'CANCEL - READY' in status:
        logger.info(f"Deleting instance: {si_uuid}")
        time.sleep(random.randint(5, 30))
        ret = workflow_api.instance_delete(si_uuid=si_uuid)
        logger.info(f"Deleted instance: {si_uuid}: ret={ret}")
    else:
        raise SenseException(f'cancel operation disrupted - instance not deleted - contact admin. {status}')


def instance_get_status(*, client=None, si_uuid):
    client = client or get_client()
    workflow_api = WorkflowCombinedApi(req_wrapper=client)
    return workflow_api.instance_get_status(si_uuid=si_uuid)


def service_instance_details(*, client=None, si_uuid, alias):
    client = client or get_client()
    discover_api = DiscoverApi(req_wrapper=client)
    response = discover_api.discover_service_instances_get(search=alias)
    # print(json.dumps(json.loads(response), indent=2))
    # response = json.loads(response)
    instances = response['instances']

    for instance in instances:
        temp = SimpleNamespace(**instance)

        if temp.referenceUUID == si_uuid:
            instance['intents'] = []

            for intent in temp.intents:
                intent['json'] = json.loads(intent['json'])
                instance['intents'].append(intent)

            return instance

    raise SenseException('no details found')


def discover_service_instances(*, client=None):
    client = client or get_client()
    discover_api = DiscoverApi(req_wrapper=client)
    response = discover_api.discover_service_instances_get()
    # print(json.dumps(json.loads(response), indent=2)) 
    # response = json.loads(response) # apparently sense client returns dicts now
    instances = response['instances']

    for instance in instances:
        temp = SimpleNamespace(**instance)
        instance['intents'] = []

        for intent in temp.intents:
            intent['json'] = json.loads(intent['json'])
            instance['intents'].append(intent)

    return instances


def find_instance_by_alias(*, client=None, alias):
    client = client or get_client()
    discover_api = DiscoverApi(req_wrapper=client)
    response = discover_api.discover_service_instances_get(search=alias)
    #response = json.loads(response) # apparently sense client returns dicts now
    instances = response['instances']

    if not instances:
        return None

    instance = instances[0]
    return instance['referenceUUID']

def old_find_instance_by_alias(*, client=None, alias):
    client = client or get_client()
    instances = discover_service_instances(client=client)

    for instance in instances:
        if instance['alias'] == alias:
            return instance['referenceUUID']

    return None


def get_vms_specs_from_profile(*, client=None, profile_uuid):
    client = client or get_client()
    profile_details = describe_profile(client=client, uuid=profile_uuid)
    all_vms = []

    if hasattr(profile_details, "intent"):
        for subnet in profile_details.intent.data.subnets:
            vms = [vars(vm) for vm in subnet.vms]
            all_vms.extend(vms)

    return all_vms


def manifest_create(*, client=None, template_file=None, alias=None, si_uuid=None):
    import os
    import time
    from json.decoder import JSONDecodeError

    client = client or get_client()
    si_uuid = si_uuid or find_instance_by_alias(alias=alias)
    workflow_api = WorkflowCombinedApi(req_wrapper=client)
    template_file = os.path.join(os.path.dirname(__file__), 'manifests', template_file)

    with open(template_file, 'r') as fp:
        template = json.load(fp)

    template = json.dumps(template)

    for attempt in range(SENSE_RETRY):
        response = workflow_api.manifest_create(template, si_uuid=si_uuid)

        try:
            response = json.loads(response, object_hook=lambda dct: SimpleNamespace(**dct))
            print (response)
            #details = json.loads(response.jsonTemplate)
            return response.get('jsonTemplate')
        except JSONDecodeError:
            logger.warning(f"Could not decode sense manifest from response={response}")

        time.sleep(10)

    raise SenseException(f"Unable to retrieve manifest using {template_file}")



from amscrot.util.utils import get_logger
import os
import logging

from amscrot.model import Service
from amscrot.provider.api.provider import Provider
from amscrot.util.constants import Constants
from amscrot.util.utils import get_inventory_dir
from amscrot.provider.janus.util.ansible_helper import AnsibleRunnerHelper


JANUS_CTRL_PORT = 5000


class JanusService(Service):
    def __init__(self, *, label, name: str, image, nodes, controller_url, controller_host,
                 controller_web, ssh_tunnel_cmd, provider, logger: logging.Logger):
        super().__init__(label=label, name=name)
        self.logger = logger
        self.image = image
        self.created = False
        self._nodes = nodes
        self._provider = provider

        if controller_url and controller_host:
            self.controller_url = controller_url
            self.controller_host = controller_host
            self._internal_controller = True
        else:
            self.controller_url = provider.config.get("url")
            self._internal_controller = False
        self.controller_web = controller_web
        self.controller_ssh_tunnel_cmd = ssh_tunnel_cmd

    def _do_ansible(self, delete=False):
        def _helper(host_file, tags, extra_vars=dict(), limit=""):
            script_dir = os.path.dirname(__file__)
            helper = AnsibleRunnerHelper(host_file, self.logger)
            helper.set_extra_vars(extra_vars)
            helper.run_playbook(os.path.join(script_dir, "ansible/janus.yml"), tags=tags, limit=limit)

        friendly_name = self._provider.name
        host_file = get_inventory_dir(friendly_name)
        janus_vars = self._provider.config
        janus_vars['url'] = self.controller_url
        if delete:
            _helper(host_file, ["janus-del"], janus_vars)
        else:
            _helper(host_file, ["docker", "janus"], janus_vars)
            if self._internal_controller:
                _helper(host_file, ["controller"], janus_vars, limit=self.controller_host)
            _helper(host_file, ["janus-add"], janus_vars)
            self.created = True

    def create(self):
        self._do_ansible()
        self.logger.info(f" Service {self.name} created. service_nodes={self._nodes}")

    def delete(self):
        self._do_ansible(delete=True)
        self.logger.info(f" Service {self.name} deleted. service_nodes={self._nodes}")


logger: logging.Logger = get_logger()


class JanusProvider(Provider):
    def setup_environment(self):
        pass

    def __init__(self, *, type, label, name,  config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        credential_file = self.config.get(Constants.CREDENTIAL_FILE)

        if credential_file:
            from amscrot.util import utils

            profile = self.config.get(Constants.PROFILE)
            config = utils.load_yaml_from_file(credential_file)
            self.config = config[profile]

    def _validate_resource(self, resource: dict):
        assert resource.get(Constants.LABEL)
        assert resource.get(Constants.RES_TYPE) in Constants.RES_SUPPORTED_TYPES
        assert resource.get(Constants.RES_NAME_PREFIX)
        creation_details = resource[Constants.RES_CREATION_DETAILS]

        # count was set to zero
        if not creation_details['in_config_file']:
            return

        assert resource.get(Constants.RES_COUNT, 1)
        self.logger.info(f"Validated:OK Resource={self.name} using {self.label}")

    def do_add_resource(self, *, resource: dict):
        """
        Called by add_resource(self, *, resource: dict) if resource has no external dependencies.
        The add_resource(self, *, resource: dict) puts resources in the pending dictionary when
        they have external dependencies.

        When the external dependencies are satisfied following a resource creation event, this method
        would be called automatically. See on_created(self, *, source, provider, resource: object)

        Note that external dependencies are respurce dependencies across different providers.
        @param resource: resource attributes
        """
        creation_details = resource[Constants.RES_CREATION_DETAILS]

        if not creation_details['in_config_file']:
            return

        self.logger.info(f"Adding resource={self.name} using {self.label}")
        self._validate_resource(resource)

        label = resource.get(Constants.LABEL)
        image = resource.get(Constants.RES_IMAGE, "dtnaas/tools")
        controller = resource.get("controller", None)
        controller_url = None
        controller_host = None
        controller_web = None
        ssh_tunnel_cmd = None
        if controller and len(controller) == 1:
            if isinstance(controller, list):
                controller = controller[0]
            if isinstance(controller, tuple):
                controller = controller[0]
            dplane_addr = controller.get_dataplane_address()
            if not dplane_addr:
                dplane_addr = "localhost"
            controller_url = f"https://{dplane_addr}:{JANUS_CTRL_PORT}"
            controller_host = controller.mgmt_ip
            controller_web = "http://localhost:8000"
            ssh_tunnel_cmd = f"{controller.sshcmd_str} -L 8000:localhost:8000"
        elif controller:
            self.logger.error(f"Invalid controller configuration for {label}")

        nodes = [rd for rd in resource[Constants.RESOLVED_EXTERNAL_DEPENDENCIES]
                 if rd.attr == 'node']
        service_name_prefix = resource.get(Constants.RES_NAME_PREFIX)
        service_nodes = [n for i in nodes for n in i.value]
        service_name = f"{self.name}-{service_name_prefix}"
        service = JanusService(label=label, name=service_name, image=image, nodes=service_nodes,
                               controller_url=controller_url,
                               controller_host=controller_host,
                               controller_web=controller_web,
                               ssh_tunnel_cmd=ssh_tunnel_cmd,
                               provider=self, logger=self.logger)
        self._services.append(service)
        self.resource_listener.on_added(source=self, provider=self, resource=service)

    def do_create_resource(self, *, resource: dict):
        """
        Called by add_resource(self, *, resource: dict) if resource has no external dependencies
        @param resource: resource attributes
        """
        creation_details = resource[Constants.RES_CREATION_DETAILS]

        if not creation_details['in_config_file']:
            return

        label = resource.get(Constants.LABEL)
        states = resource.get(Constants.SAVED_STATES)
        force = resource.get("force", False)
        created = False
        for s in states:
            created = s.attributes.get('created', True)

        self.logger.info(f"Creating resource={self.name} using {self.label}")

        temp = [service for service in self.services if service.label == label]

        for service in temp:
            if not force and created:
                self.logger.info(f"Service {label} is already in created state, skipping create task")
                service.created = True
            else:
                service.create()
            self.resource_listener.on_created(source=self, provider=self, resource=service)

    def do_delete_resource(self, *, resource: dict):
        self.logger.info(f"Deleting resource={resource} using {self.label}")
        image = resource.get(Constants.RES_IMAGE, "dtnaas/tools")
        nodes = list()

        if resource.get(Constants.EXTERNAL_DEPENDENCY_STATES):
            nodes = [rd for rd in resource.get(Constants.EXTERNAL_DEPENDENCY_STATES)]

        service_nodes = [n.attributes.get('name') for n in nodes]
        label = resource.get(Constants.LABEL)
        states = resource.get(Constants.SAVED_STATES)
        states = [s for s in states if s.label == label]

        if not states:
            return

        s = states[0]
        controller_url = s.attributes.get('controller_url')
        controller_host = s.attributes.get('controller_host')
        controller_web = s.attributes.get('controller_web')
        ssh_tunnel_cmd = s.attributes.get('controller_ssh_tunnel_cmd')

        service_name_prefix = resource.get(Constants.RES_NAME_PREFIX)
        service_name = f"{self.name}-{service_name_prefix}"
        service = JanusService(label=label, name=service_name, image=image, nodes=service_nodes,
                               controller_url=controller_url,
                               controller_host=controller_host,
                               controller_web=controller_web,
                               ssh_tunnel_cmd=ssh_tunnel_cmd,
                               provider=self, logger=self.logger)

        service.delete()
        self.resource_listener.on_deleted(source=self, provider=self, resource=service)



import logging

from amscrot.provider.api.provider import Provider
from ...util.constants import Constants
from .fabric_constants import *
from amscrot.util.utils import get_logger
from amscrot.exceptions import ProviderException

logger = get_logger()


class FabricProvider(Provider):
    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        self.slice = None
        self.retry = 5
        self.slice_init = False
        # TODO Should not be needed for fablib 1.6.4
        from fabrictestbed_extensions.fablib.constants import Constants as FC
        from amscrot.util.utils import get_base_dir

        FC.DEFAULT_FABRIC_CONFIG_DIR = get_base_dir(self.name)
        FC.DEFAULT_FABRIC_RC = f"{FC.DEFAULT_FABRIC_CONFIG_DIR}/fabric_rc"



    def _to_abs_for(self, env_var: str, config: dict):
        path = config.get(env_var)

        from pathlib import Path

        path = str(Path(path).expanduser().absolute())

        try:
            with open(path, 'r') as f:
                f.read()
                logger.debug(f"was able to read {path}")
        except Exception as e:
            raise ProviderException(f"{self.name}:Unable to read {path} for {env_var}:{e}")

        try:
            if path.endswith(".json"):
                with open(path, 'r') as fp:
                    import json

                    json.load(fp)
        except Exception as e:
            raise ProviderException(f"{self.name}:Unable to parse {path} to json for {env_var}:{e}")

        return path

    def setup_environment(self):
        config = self.config

        import os

        for attr in FAB_CONF_ATTRS:
            if config.get(attr) is None:
                raise ProviderException(f"{self.name}: Expecting a value for {attr}")

        os.environ['FABRIC_CREDMGR_HOST'] = config.get(FABRIC_CM_HOST, DEFAULT_CM_HOST)
        os.environ['FABRIC_ORCHESTRATOR_HOST'] = config.get(FABRIC_OC_HOST, DEFAULT_OC_HOST)
        os.environ['FABRIC_PROJECT_ID'] = config.get(FABRIC_PROJECT_ID)
        os.environ['FABRIC_BASTION_HOST'] = config.get(FABRIC_BASTION_HOST, DEFAULT_BASTION_HOST)
        os.environ['FABRIC_BASTION_USERNAME'] = config.get(FABRIC_BASTION_USER_NAME)

        os.environ['FABRIC_BASTION_KEY_LOCATION'] = self._to_abs_for(FABRIC_BASTION_KEY_LOCATION, config)
        os.environ['FABRIC_SLICE_PRIVATE_KEY_FILE'] = self._to_abs_for(FABRIC_SLICE_PRIVATE_KEY_LOCATION, config)
        os.environ['FABRIC_SLICE_PUBLIC_KEY_FILE'] = self._to_abs_for(FABRIC_SLICE_PUBLIC_KEY_LOCATION, config)

        token_location = self._to_abs_for(FABRIC_TOKEN_LOCATION, config)

        if Constants.COPY_TOKENS:
            import shutil
            import uuid

            destination = f'/tmp/tokens/token-{self.name}-{uuid.uuid4()}.json'
            shutil.copy2(token_location, destination)
            token_location = destination

        os.environ['FABRIC_TOKEN_LOCATION'] = token_location

        from . import fabric_slice_helper

        fabric_slice_helper.patch_for_token()

    def _init_slice(self, destroy_phase=False):
        if not self.slice_init:
            self.logger.info(f"Initializing slice {self.name}")

            import time
            from amscrot.util.utils import get_log_level, get_log_location

            location = get_log_location()

            for attempt in range(self.retry):
                try:
                    from fabrictestbed_extensions.fablib.fablib import fablib

                    if fablib.get_default_fablib_manager().get_log_file() != location:
                        self.logger.debug("Initializing fablib extensions logging ...")
                        fablib.get_default_fablib_manager().set_log_file(location)
                        fablib.get_default_fablib_manager().set_log_level(get_log_level())

                        for handler in logging.root.handlers.copy():
                            logging.root.removeHandler(handler)

                        for handler in self.logger.handlers:
                            logging.root.addHandler(handler)

                    from amscrot.provider.fabric.fabric_slice import FabricSlice

                    temp = FabricSlice(provider=self, logger=self.logger)
                    temp.init(destroy_phase)
                    self.slice = temp
                    self.logger.info(f"Done initializing slice {self.name}")
                    break
                except Exception as e:
                    if attempt == self.retry - 1:
                        raise e

                    self.logger.info(f"Initializing slice {self.name}. Going to sleep. Will retry ...{e}")
                time.sleep(2)

            self.logger.info(f"Initialized slice {self.name}")
            self.slice_init = True

    def supports_modify(self):
        return True

    def do_validate_resource(self, *, resource: dict):
        self._init_slice()
        assert self.slice.slice_object is not None
        self.slice.validate_resource(resource=resource)

    def do_add_resource(self, *, resource: dict):
        assert self.slice.slice_object is not None
        self.slice.add_resource(resource=resource)

    def do_create_resource(self, *, resource: dict):
        assert self.slice.slice_object is not None
        self.slice.create_resource(resource=resource)

    def do_wait_for_create_resource(self, *, resource: dict):
        assert self.slice.slice_object is not None
        self.slice.wait_for_create_resource(resource=resource)

    def do_delete_resource(self, *, resource: dict):
        self._init_slice(True)
        self.slice.delete_resource(resource=resource)



import unittest
import json
import logging
from amscrot.serviceclient.kube.kube_service_client import KubeServiceClient
from amscrot.serviceclient.esnet_iri.esnet_iri_service_client import EsnetIriServiceClient
from amscrot.serviceclient.iri.iri_service_client import IriServiceClient
from amscrot.model.discovery import DiscoveryResult
from amscrot.util.constants import Constants

# Configure logging to show output during tests (if -s is used)
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class TestServiceClientDiscovery(unittest.TestCase):
    """Test the discover method of ServiceClient implementations."""

    def test_kube_discover(self):
        """Test KubeServiceClient.discover().
        
        Skips if Kubernetes client is not available or connection fails.
        Dumps discovery JSON to stdout on success.
        """
        print("\n--- Testing KubeServiceClient.discover ---")
        client = KubeServiceClient(name="test-kube")
        
        if not client._available:
             self.skipTest("KubeServiceClient unavailable (no kubeconfig found or load failed).")

        try:
            discovery = client.discover()
            
            # Basic assertions
            self.assertIsInstance(discovery, DiscoveryResult)
            
            nodes = discovery.by_type('node')
            crds = discovery.by_type('crd')
            
            print(f"Discovery Summary:")
            print(f"  Total items: {len(discovery)}")
            print(f"  Nodes found: {len(nodes)}")
            print(f"  CRDs found: {len(crds)}")
            print(f"  Types: {discovery.summary()}")
            
            if not discovery:
                 self.skipTest("KubeServiceClient likely connected but returned no resources (empty cluster?).")

            # Dump complete JSON
            print("\n--- Discovery JSON Dump ---")
            print(json.dumps(discovery.to_list(), indent=2, default=str))
            print("--- End JSON Dump ---")
            
        except Exception as e:
            # Re-raise SkipTest so it's not treated as a failure
            if isinstance(e, unittest.SkipTest):
                raise

            # Check for connection errors and skip if appropriate
            error_msg = str(e)
            if "Max retries exceeded" in error_msg or "Connection refused" in error_msg: 
                 self.skipTest(f"Connection to Kubernetes cluster failed: {e}")
            else:
                 self.fail(f"KubeServiceClient discovery failed with unexpected error: {e}")

    def test_esnet_iri_discover(self):
        """Test EsnetIriServiceClient.discover()."""
        print("\n--- Testing EsnetIriServiceClient.discover ---")
        # Use a name that matches an entry in credentials.yml
        client = EsnetIriServiceClient(name="esnet-iri-west", profile="esnet-iri-west")
        # Discovery is a stub, so it should always succeed returned empty list
        try:
            discovery = client.discover()
            
            # If no credentials, it returns empty result (which is valid behavior for client itself, 
            # though locally we expect credentials to be present now).
            if not discovery:
                 print("EsnetIriServiceClient returned empty discovery. Check credentials if this is unexpected.")
                 return

            # Basic assertions for extended discovery
            self.assertIsInstance(discovery, DiscoveryResult)
            
            print(f"Esnet IRI Discovery Summary:")
            print(f"  Total items: {len(discovery)}")
            print(f"  Compute resources: {len(discovery.compute)}")
            print(f"  Facilities: {len(discovery.facility)}")
            print(f"  Capabilities: {len(discovery.capability)}")
            print(f"  Allocations: {len(discovery.allocation)}")
            print(f"  Types: {discovery.summary()}")

            # We expect at least some of these if connected to a real environment
            if not (discovery.compute or discovery.facility or discovery.capability or discovery.allocation):
                 self.skipTest("EsnetIriServiceClient connected but found no resources of any type.")

            # Dump JSON for manual inspection
            print("\n--- Esnet IRI Discovery JSON Dump ---")
            print(json.dumps(discovery.to_list(), indent=2, default=str))
            print("--- End JSON Dump ---")

        except Exception as e:
            # Handle potential credential load errors if they propagate
             self.fail(f"EsnetIriServiceClient discovery failed: {e}")

    def test_iri_discover(self):
        """Test IriServiceClient.discover()."""
        print("\n--- Testing IriServiceClient.discover ---")
        client = IriServiceClient(name="test-iri")
        # Discovery is a stub
        discovery = client.discover()
        self.assertIsInstance(discovery, DiscoveryResult)
        self.assertEqual(len(discovery), 0, "IriServiceClient.discover should return empty DiscoveryResult")

if __name__ == '__main__':
    unittest.main()




import unittest
from amscrot.client import Job, JobSpec, JobType, JobServiceType, JobState
from amscrot.serviceclient import ServiceClient
from amscrot.util.constants import Constants

class TestClientJob(unittest.TestCase):
    def test_service_client_creation(self):
        sc = ServiceClient.create(type=Constants.ServiceType.IRI, name="sc1", endpoint_uri="http://localhost:8000")
        self.assertEqual(sc.name, "sc1")
        self.assertEqual(sc.endpoint_uri, "http://localhost:8000")
        self.assertEqual(sc.status().state, JobState.ACTIVE)
        self.assertEqual(sc.capabilities, [])
        
    def test_job_creation(self):
        sc = ServiceClient.create(type=Constants.ServiceType.IRI, name="sc1", endpoint_uri="http://localhost:8000")
        
        spec = JobSpec(
            resources={"node_count": 10},
            image="my-image",
            executable=["run.sh"],
            attributes={"duration": 60}
        )
        
        job = Job(
            name="job1",
            type=JobType.COMPUTE,
            service_type=JobServiceType.BATCH,
            service_client=sc,
            job_spec=spec,
            preferences={"site": "nersc"}
        )
        
        self.assertEqual(job.name, "job1")
        self.assertEqual(job.type, JobType.COMPUTE)
        self.assertEqual(job.service_type, JobServiceType.BATCH)
        self.assertEqual(job.status, JobState.INIT)
        self.assertEqual(job.service_client, sc)
        self.assertEqual(job.job_spec, spec)
        self.assertEqual(job.preferences['site'], 'nersc')
        
    def test_job_enum_conversion(self):
        # Test passing strings instead of Enums
        job = Job(
            name="job_str",
            type="DATA",
            service_type="REALTIME"
        )
        self.assertEqual(job.type, JobType.DATA)
        self.assertEqual(job.service_type, JobServiceType.REALTIME)
        
    def test_job_dependency(self):
        job1 = Job(name="j1", type="COMPUTE", service_type="BATCH")
        job2 = Job(name="j2", type="DATA", service_type="REALTIME", dependency=job1)
        
        self.assertEqual(job2.dependency, job1)

    def test_job_spec_defaults(self):
        spec = JobSpec()
        self.assertEqual(spec.resources, {})
        self.assertEqual(spec.executable, [])

    def test_add_job_to_session(self):
        from amscrot.client import Client
        
        client = Client()
        session = client.create_session("job_session")
        
        job = Job(name="j1", type="COMPUTE", service_type="BATCH")
        session.add_job(job)
        
        config = session._build_config()
        self.assertEqual(len(config['job']), 1)
        
        # config['job'] is a list of dicts returned by Job.to_config()
        # e.g. [{'j1': ...}]
        job_config = config['job'][0]['j1']
        self.assertEqual(job_config['name'], "j1")
        self.assertEqual(job_config['type'], "COMPUTE")



from types import SimpleNamespace
from typing import List, Tuple, Union

from .config_models import *
from .constants import Constants
from .resource_dependency_helper import order_resources
from .variable_evaluator import VariableEvaluator, Evaluator


def normalize(objs):  # TODO looks like we need this  for chameleon provider
    for obj in objs:
        attrs = {}

        for key, value in obj.attributes.items():
            if isinstance(value, SimpleNamespace):
                attrs[key] = vars(value)
            else:
                attrs[key] = value

        obj.attributes = attrs


def parse_pair(obj: SimpleNamespace) -> Tuple[str, Dict]:
    def extract_label(lst):
        return next(label for label in lst if not label.startswith("__"))

    if not isinstance(obj, SimpleNamespace):
        raise ParseConfigException(f"expecting a pair {obj}")

    try:
        name = extract_label(dir(obj))

        if not obj.__getattribute__(name):
            return name, {}

        if isinstance(obj.__getattribute__(name), list):
            attrs = obj.__getattribute__(name)[0].__dict__

            if len(obj.__getattribute__(name)) > 1:
                raise ParseConfigException(f"did not expect a block after {attrs} under {name} of type {type}")
        else:
            attrs = obj.__getattribute__(name).__dict__

        return name, attrs
    except ParseConfigException as e:
        raise e
    # except Exception as e:
    #     raise ParseConfigException(f" exception occurred while parsing pair {obj}") from e


def parse_triplet(obj: SimpleNamespace) -> List[Tuple[str, str, Dict]]:
    def extract_label(lst):
        return next(label for label in lst if not label.startswith("__"))

    if not isinstance(obj, SimpleNamespace):
        raise ParseConfigException(f"expecting a triplet {obj}")

    try:
        type = extract_label(dir(obj))
        triplets = []

        for v in obj.__getattribute__(type):
            name = extract_label(dir(v))

            if v.__getattribute__(name) is None:
                attrs = dict()
            elif isinstance(v.__getattribute__(name), list):
                attrs = v.__getattribute__(name)[0].__dict__

                if len(v.__getattribute__(name)) > 1:
                    raise ParseConfigException(f"did not expect a block after {attrs} under {name} of type {type}")
            else:
                attrs = v.__getattribute__(name).__dict__

            triplets.append((type, name, attrs))

        return triplets
    except ParseConfigException as e:
        raise e
    except Exception as e:
        raise ParseConfigException(f" exception occurred while parsing triplet {obj}") from e


class Parser:
    def __init__(self):
        pass

    @staticmethod
    def _parse_variable(obj) -> Variable:
        name, attributes = parse_pair(obj)
        return Variable(name, attributes.get('default', None))

    @staticmethod
    def _parse_provider(obj) -> List[ProviderConfig]:
        return [ProviderConfig(type, name, attributes) for type, name, attributes in parse_triplet(obj)]

    @staticmethod
    def _parse_config(obj) -> List[Config]:
        return [Config(type, name, attributes) for type, name, attributes in parse_triplet(obj)]

    @staticmethod
    def _parse_resource_base_config(obj) -> List[BaseConfig]:
        return [BaseConfig(type, name, attributes) for type, name, attributes in parse_triplet(obj)]

    @staticmethod
    def _filter_resources(base_configs, providers) -> List[ResourceConfig]:
        resources = []

        for base_config in base_configs:
            resources.append(resource_from_basic_config(base_config, providers))

        return resources

    @staticmethod
    def _validate_variables(variables: List[Variable]):
        if len(variables) != len(set(variables)):
            raise ParseConfigException(f'detected duplicate variables')

        for variable in variables:
            if variable.value is None:
                raise ParseConfigException(f'variable {variable .name} is not bound')

    @staticmethod
    def _validate_providers(providers: List[ProviderConfig]):
        # if len(providers) == 0:
        #     raise ParseConfigException("no providers found ...")

        if len(providers) != len(set(providers)):
            raise ParseConfigException(f'detected duplicate providers')

        from amscrot.exceptions import ProviderTypeNotSupported

        for provider in providers:
            if provider.type not in Constants.PROVIDER_CLASSES:
                raise ProviderTypeNotSupported(provider.type)

    @staticmethod
    def _validate_configs(configs: List[Config]):
        from amscrot.exceptions import ConfigTypeNotSupported

        for config in configs:
            if config.type not in Constants.CONFIG_SUPPORTED_TYPES:
                raise ConfigTypeNotSupported(config.type)

    @staticmethod
    def _validate_resources(resources: List[ResourceConfig]):
        # if len(resources) == 0:
        #     raise ParseConfigException("no resources found ...")

        if len(resources) != len(set(resources)):
            raise ParseConfigException(f'detected duplicate resources')

        from amscrot.exceptions import ResourceTypeNotSupported

        for resource in resources:
            if resource.type not in Constants.RES_SUPPORTED_TYPES:
                raise ResourceTypeNotSupported(resource.type)

        for resource in resources:
            count = resource.attributes.get(Constants.RES_COUNT, 1)

            if count > 1 and resource.type in Constants.RES_RESTRICTED_TYPES:
                raise ParseConfigException(f'{resource.label} is of type {resource.type} and cannot have count > 1')

    @staticmethod
    def parse_variables(ns_list: list, var_dict: dict) -> List[Variable]:
        variables = []

        for ns in ns_list:
            if hasattr(ns, 'variable') and ns.variable:
                temp_variables = [Parser._parse_variable(variable) for variable in ns.variable]
                variables.extend(temp_variables)

        if var_dict:
            variable_map = {v.name: v for v in variables}

            for key, value in var_dict.items():
                variable_map[key] = Variable(key, value)

            variables = list(variable_map.values())

        Parser._validate_variables(variables)
        return variables

    @staticmethod
    def parse_providers(ns_list: list) -> List[ProviderConfig]:
        providers = []

        for ns in ns_list:
            if hasattr(ns, 'provider') and ns.provider:
                temp_providers = [Parser._parse_provider(provider) for provider in ns.provider]
                temp_providers = sum(temp_providers, [])
                providers.extend(temp_providers)

        Parser._validate_providers(providers)
        normalize(providers)
        return providers

    @staticmethod
    def parse_configs(ns_list: List[SimpleNamespace]) -> List[Config]:
        configs = []

        for ns in ns_list:
            if hasattr(ns, 'config') and ns.config:
                temp_configs = [Parser._parse_config(config) for config in ns.config]
                temp_configs = sum(temp_configs, [])
                configs.extend(temp_configs)

        Parser._validate_configs(configs)
        return configs

    @staticmethod
    def parse_resource_base_configs(ns_list: List[SimpleNamespace]) -> List[BaseConfig]:
        resource_base_configs = []

        for ns in ns_list:
            if hasattr(ns, 'resource') and ns.resource:
                temp_resource_base_configs = [Parser._parse_resource_base_config(resource) for resource in ns.resource]
                temp_resource_base_configs = sum(temp_resource_base_configs, [])
                resource_base_configs.extend(temp_resource_base_configs)

        return resource_base_configs

    @staticmethod
    def parse(*, dir_path: Union[str, None] = None, content: Union[str, None] = None,
              var_dict: Union[Dict, None] = None) -> Tuple[List[ProviderConfig], List[ResourceConfig]]:

        from .utils import load_as_ns_from_yaml

        ns_list = load_as_ns_from_yaml(dir_path=dir_path, content=content)
        variables = Parser.parse_variables(ns_list, var_dict)

        providers = Parser.parse_providers(ns_list)
        configs = Parser.parse_configs(ns_list)
        resource_base_configs = Parser.parse_resource_base_configs(ns_list)
        variable_evaluator = VariableEvaluator(variables=variables, providers=providers, configs=configs,
                                               resources=resource_base_configs)
        providers, configs, resource_configs = variable_evaluator.evaluate()

        evaluator = Evaluator(providers=providers, configs=configs, resources=resource_configs)

        providers, configs, resource_configs = evaluator.evaluate()
        resources = Parser._filter_resources(resource_configs, providers)

        Parser._validate_resources(resources)
        resources = [r for r in resources if r.attributes.get(Constants.RES_COUNT, 1) > 0]

        from .resource_dependency_helper import ResourceDependencyEvaluator

        dependency_evaluator = ResourceDependencyEvaluator(resources, providers)
        dependency_map = dependency_evaluator.evaluate()
        ordered_resources = order_resources(dependency_map)
        return providers, ordered_resources



import time

from google.cloud import compute_v1
from google.cloud.compute_v1.types import (
    Router,
    RouterBgp,
    # RouterBgpPeer,
    # RouterBgpPeerBfd,
    RouterMd5AuthenticationKey,
    InsertRouterRequest,
    PatchRouterRequest,
    InterconnectAttachment,
    Operation
)
from google.oauth2 import service_account

from amscrot.util.utils import get_logger

logger = get_logger()


GCP_REQUEST_RETRY_MAX = 10 


def find_vpc(*, service_key_path, project, vpc):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    network_client = compute_v1.NetworksClient(credentials=credentials)
    request = compute_v1.GetNetworkRequest(project=project, network=vpc)

    from google.api_core.exceptions import NotFound

    try:
        return network_client.get(request=request)
    except NotFound:
        return None


def check_operation_result(*, credentials, project, region, operation_name):
    client = compute_v1.RegionOperationsClient(credentials=credentials)
    request = compute_v1.GetRegionOperationRequest(
        operation=operation_name,
        project=project,
        region=region,
    )

    for attemp in range(GCP_REQUEST_RETRY_MAX):
        response = client.get(request=request)

        if response.status == Operation.Status.DONE:
            return
        else:
            logger.info(f"Operation {operation_name} not done: Status={response.status}. Retrying in 5 seconds...")
            time.sleep(5)           
    
    raise Exception(f"Operation {operation_name} failed after maximum retries {GCP_REQUEST_RETRY_MAX}.")


def find_router(*, service_key_path, project, region, router_name):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    compute_client = compute_v1.RoutersClient(credentials=credentials)
    request = compute_v1.GetRouterRequest(
                    project=project,
                    region=region,
                    router=router_name
    )

    from google.api_core.exceptions import NotFound

    try:
        return compute_client.get(request=request)
    except NotFound:
        return None


def create_router(*, service_key_path, project, region, router_name, vpc, bgp_asn):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    compute_client = compute_v1.RoutersClient(credentials=credentials)
    router_resource = Router(
        name=router_name,
        network=f'projects/{project}/global/networks/{vpc}',
        bgp=RouterBgp(asn=bgp_asn)
    )
    request = InsertRouterRequest(
        project=project,
        region=region,
        router_resource=router_resource
    )
    response = compute_client.insert(request=request)
    logger.info(f'Response={response}')
    check_operation_result(credentials=credentials, project=project, region=region, operation_name=response.name)
    logger.info(f'Router {router_name} created successfully.')


def patch_router(*, service_key_path, project, region, router_name, bgp_key):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    compute_client = compute_v1.RoutersClient(credentials=credentials)
    
    # Get the router resource
    router = compute_client.get(project=project, region=region, router=router_name)
    
    # Access the BGP peers associated with the router
    bgp_peers = router.bgp_peers
    
    # Add md5 authentication
    peer = bgp_peers[0]
    peer.md5_authentication_key_name = 'md5-key-name-1'
    
    router_resource = Router(
        name=router_name,
        md5_authentication_keys=[RouterMd5AuthenticationKey(key=bgp_key, name='md5-key-name-1')],
        bgp_peers=[peer]
    )
    request = PatchRouterRequest(
        project=project,
        region=region,
        router=router_name,
        router_resource=router_resource)
    response = compute_client.patch(request=request)
    logger.info(f'Response={response}')
    check_operation_result(credentials=credentials, project=project, region=region, operation_name=response.name)
    logger.info(f'Router {router_name} patched successfully.')


def delete_router(*, service_key_path, project, region, router_name):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    client = compute_v1.RoutersClient(credentials=credentials)
    request = compute_v1.DeleteRouterRequest(
        project=project,
        region=region,
        router=router_name,
    )
    
    response = client.delete(request=request)
    check_operation_result(credentials=credentials, project=project,
                           region=region, operation_name=response.name)
    logger.info(f"Router '{router_name}' deleted successfully.")


def find_interconnect_attachment(*, service_key_path, project, region, attachment_name):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    compute_client = compute_v1.InterconnectAttachmentsClient(credentials=credentials)

    request = compute_v1.GetInterconnectAttachmentRequest(
                    project=project,
                    region=region,
                    interconnect_attachment=attachment_name
    )

    from google.api_core.exceptions import NotFound

    try:
        return compute_client.get(request=request)
    except NotFound:
        return None


def create_interconnect_attachment(*, service_key_path, project, region, mtu, router_name, attachment_name):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    compute_client = compute_v1.InterconnectAttachmentsClient(credentials=credentials)
    interconnect_attachment_resource = InterconnectAttachment(
        name=f'{attachment_name}',
        admin_enabled=True,
        encryption=None,
        router=f'projects/{project}/regions/{region}/routers/{router_name}',
        type_='PARTNER',
        mtu=mtu
    )
    
    request = compute_v1.InsertInterconnectAttachmentRequest(
        project=project,
        region=region,
        interconnect_attachment_resource=interconnect_attachment_resource
    )
    response = compute_client.insert(request=request)
    check_operation_result(credentials=credentials, project=project, region=region, operation_name=response.name)
    logger.info(f"Interconnect VLAN Attachment '{attachment_name}' created successfully.")


def delete_interconnect_vlan_attachment(*, service_key_path, project, region, attachment_name):
    credentials = service_account.Credentials.from_service_account_file(service_key_path)
    client = compute_v1.InterconnectAttachmentsClient(credentials=credentials)
    request = compute_v1.DeleteInterconnectAttachmentRequest(
        project=project,
        region=region,
        interconnect_attachment=attachment_name,
    )

    response = client.delete(request=request)
    check_operation_result(credentials=credentials, project=project, region=region, operation_name=response.name)
    logger.info(f"Interconnect VLAN Attachment '{attachment_name}' deleted successfully.")



ACCESS_KEY = "ACCESS_KEY"
SECRET_KEY = "SECRET_KEY"
RETRY = 60

VLAN = 'vlan'
AMAZON_SIDE_ASN = 'amazonSideAsn'
ASN = 'asn'
VIF_ID = 'virtualInterfaceId'
VIF_STATE = 'virtualInterfaceState'

VIF_DETAILS = [VLAN, AMAZON_SIDE_ASN, ASN, VIF_ID, VIF_STATE]
DO_NOT_USE_MAIN_ROUTING_TABLE = True



import unittest
import pytest
from amscrot.client import Client, Session, Job, JobType, JobServiceType, JobSpec, JobState
from amscrot.serviceclient import ServiceClient
from amscrot.util.constants import Constants

class TestClientJobsLifecycle(unittest.TestCase):
    def test_jobs_lifecycle(self):
        client = Client()
        
        session = client.create_session("test_jobs_lifecycle")
        
        # Create ServiceClients
        sc1 = ServiceClient.create(type=Constants.ServiceType.IRI,
                                   name="nersc_perlmutter",
                                   endpoint_uri="https://nersc.gov/api")
        sc2 = ServiceClient.create(type=Constants.ServiceType.DUMMY,
                                   name="anl_theta")
        
        # Job 1 on SC1
        job1 = Job(
            name="job1", 
            type=JobType.COMPUTE, 
            service_type=JobServiceType.BATCH,
            service_client=sc1,
            job_spec=JobSpec(
                resources={"node_count": 1},
                image="alpine:latest",
                executable=["/bin/sh", "-c", "echo hello job1"],
                attributes={"walltime": "10m"}
            )
        )
        
        # Job 2 on SC1, depends on Job1
        job2 = Job(
            name="job2", 
            type=JobType.DATA, 
            service_type=JobServiceType.BATCH,
            service_client=sc1,
            dependency=job1,
            job_spec=JobSpec(
                resources={"storage_gb": 100},
                image="data-mover:latest",
                executable=["/bin/transfer"],
                attributes={"bandwidth": "10Gbps"}
            )
        )
        
        # Job 3 on SC2
        job3 = Job(
            name="job3", 
            type=JobType.COMPUTE, 
            service_type=JobServiceType.REALTIME,
            service_client=sc2,
            job_spec=JobSpec(
                resources={"gpu_count": 4},
                image="tensorflow:latest",
                executable=["python", "train.py"],
                attributes={"priority": "high"}
            )
        )
        
        # Job 4 on SC2, depends on Job3
        job4 = Job(
            name="job4", 
            type=JobType.INSTRUMENT, 
            service_type=JobServiceType.REALTIME,
            service_client=sc2,
            dependency=job3,
             job_spec=JobSpec(
                resources={"sensor_id": "sensor-1"},
                image="instrument-ctrl:v1",
                executable=["/bin/capture"],
                attributes={"duration": "1h"}
            )
        )
        
        session.add_job(job1)
        session.add_job(job2)
        session.add_job(job3)
        session.add_job(job4)
    
        
        print("\n--- PLAN ---")
        try:
            session.plan()
        except Exception as e:
            # Skipping test if any error occurs during plan
            self.skipTest(f"Skipping test - an error occurred during plan: {e}")
        
        print("\n--- APPLY ---")
        session.apply()
        
        print("\n--- SHOW ---")
        session.show()
        
        print("\n--- DESTROY ---")
        session.destroy()
        



from typing import Dict, List

import yaml

from amscrot.util.config_models import Config, ResourceConfig, BaseConfig, ProviderConfig, Dependency, DependencyInfo
from amscrot.model import ResolvedDependency
from amscrot.util.constants import Constants


class BaseState:
    def __init__(self, type: str, label: str, attributes: Dict):
        self.type = type
        self.label = label
        self.attributes = attributes


class ResourceState(BaseState):
    def __init__(self, type: str, label: str, attributes: Dict):
        super().__init__(type, label, attributes)
        self.type = type
        self.label = label
        self.attributes = attributes

    @property
    def name(self) -> str:
        return self.attributes['name']

    @property
    def is_node_state(self):
        return self.type == Constants.RES_TYPE_NODE

    @property
    def is_network_state(self):
        return self.type == Constants.RES_TYPE_NETWORK

    @property
    def is_service_state(self):
        return self.type == Constants.RES_TYPE_SERVICE


class NetworkState(ResourceState):
    def __init__(self, *, label, attributes):
        super().__init__(Constants.RES_TYPE_NETWORK, label, attributes)


class NodeState(ResourceState):
    def __init__(self, *, label, attributes):
        super().__init__(Constants.RES_TYPE_NODE, label, attributes)


class ServiceState(ResourceState):
    def __init__(self, *, label, attributes):
        super().__init__(Constants.RES_TYPE_SERVICE, label, attributes)


class ProviderState(BaseState):
    def __init__(self, label, attributes, network_states: List[NetworkState],
                 node_states: List[NodeState], service_states: List[ServiceState], pending,
                 pending_internal, failed: Dict[str, str], creation_details: Dict):
        super().__init__("provider", label, attributes)
        self.network_states = network_states
        self.node_states = node_states
        self.service_states = service_states
        self.pending = pending
        self.pending_internal = pending_internal
        self.failed = failed
        self.creation_details = creation_details

    def add_if_not_found(self, resource_state: ResourceState):
        from typing import cast
        assert resource_state.type in [Constants.RES_TYPE_NODE,
                                       Constants.RES_TYPE_NETWORK,
                                       Constants.RES_TYPE_SERVICE]

        def exists(state, states: list):
            temp = next(filter(lambda s: s.label == state.label and s.name == state.name, states), None)

            return temp is not None

        if exists(resource_state, self.states()):
            return

        if resource_state.is_node_state:
            self.node_states.append(cast(NodeState, resource_state))
        elif resource_state.is_network_state:
            self.network_states.append(cast(NetworkState, resource_state))
        elif resource_state.is_service_state:
            self.service_states.append(cast(ServiceState, resource_state))

    def add(self, resource_state: ResourceState):
        from typing import cast
        assert resource_state.type in [Constants.RES_TYPE_NODE,
                                       Constants.RES_TYPE_NETWORK,
                                       Constants.RES_TYPE_SERVICE]

        def exists(state, states: list):
            temp = next(filter(lambda s: s.label == state.label and s.name == state.name, states), None)

            return temp is not None

        assert not exists(resource_state, self.states())

        if resource_state.is_node_state:
            self.node_states.append(cast(NodeState, resource_state))
        elif resource_state.is_network_state:
            self.network_states.append(cast(NetworkState, resource_state))
        elif resource_state.is_service_state:
            self.service_states.append(cast(ServiceState, resource_state))

    def add_all(self, resource_states: List[ResourceState]):
        for resource_state in resource_states:
            self.add(resource_state)

    def states(self) -> List[ResourceState]:
        temp = []
        temp.extend(self.network_states)
        temp.extend(self.node_states)
        temp.extend(self.service_states)
        return temp

    def number_of_created_resources(self):
        count = 0

        for creation_detail in self.creation_details.values():
            count += creation_detail['created_count']

        return count

    def number_of_failed_resources(self):
        count = 0

        for creation_detail in self.creation_details.values():
            count += creation_detail['failed_count']

        return count

    def number_of_total_resources(self):
        count = 0

        for creation_detail in self.creation_details.values():
            count += creation_detail['total_count']

        return count


def provider_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> ProviderState:
    return ProviderState(**loader.construct_mapping(node))


def provider_representer(dumper: yaml.SafeDumper, provider_state: ProviderState) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!ProviderState", {
        "label": provider_state.label,
        "attributes": provider_state.attributes,
        "network_states": provider_state.network_states,
        "node_states": provider_state.node_states,
        "service_states": provider_state.service_states,
        "pending": provider_state.pending,
        "pending_internal": provider_state.pending_internal,
        "failed": provider_state.failed,
        "creation_details": provider_state.creation_details
    })


def network_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> NetworkState:
    return NetworkState(**loader.construct_mapping(node))


def network_representer(dumper: yaml.SafeDumper, network_state: NetworkState) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!NetworkState", {
        "label": network_state.label,
        "attributes": network_state.attributes
    })


def node_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> NodeState:
    return NodeState(**loader.construct_mapping(node))


def node_representer(dumper: yaml.SafeDumper, node_state: NodeState) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!NodeState", {
        "label": node_state.label,
        "attributes": node_state.attributes
    })


def base_config_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> BaseConfig:
    return BaseConfig(**loader.construct_mapping(node))


def base_config_representer(dumper: yaml.SafeDumper, base_config: BaseConfig) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!BaseConfig", {
        "type": base_config.type,
        "name": base_config.name,
        "attrs": base_config.attributes
    })


def config_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> Config:
    return Config(**loader.construct_mapping(node))


def config_representer(dumper: yaml.SafeDumper, config: Config) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!Config", {
        "type": config.type,
        "name": config.name,
        "attrs": config.attributes
    })


def provider_config_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> ProviderConfig:
    return ProviderConfig(**loader.construct_mapping(node))


def provider_config_representer(dumper: yaml.SafeDumper, provider_config: ProviderConfig) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!ProviderConfig", {
        "type": provider_config.type,
        "name": provider_config.var_name,
        "attrs": provider_config.attributes
    })


def resource_config_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> ResourceConfig:
    return ResourceConfig(**loader.construct_mapping(node))


def resource_config_representer(dumper: yaml.SafeDumper, resource_config: ResourceConfig) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!ResourceConfig", {
        "type": resource_config.type,
        "name": resource_config.var_name,
        "attrs": resource_config.attributes,
        "provider": resource_config.provider
    })


def service_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> ServiceState:
    return ServiceState(**loader.construct_mapping(node))


def service_representer(dumper: yaml.SafeDumper, service_state: ServiceState) -> yaml.nodes.MappingNode:
    return dumper.represent_mapping("!ServiceState", {
        "label": service_state.label,
        "attributes": service_state.attributes
    })


def named_tuple(self, data):
    if hasattr(data, '_asdict'):
        if isinstance(data, DependencyInfo):
            data = dict(dependency_info=dict(resource=str(data.resource),
                                             attribute=data.attribute))
        elif isinstance(data, Dependency):
            data = dict(dependency=dict(resource=str(data.resource),
                                        key=data.key,
                                        is_external=data.is_external,
                                        attribute=data.attribute))
        elif isinstance(data, ResolvedDependency):
            data = dict(resolved_dependency=dict(resource_label=data.resource_label,
                                                 attribute=data.attr,
                                                 value=str(data.value)))
        else:
            data = data._asdict()

        return self.represent_dict(data)

    return self.represent_list(data)


def get_loader():
    loader = yaml.SafeLoader
    loader.add_constructor("!NetworkState", network_constructor)
    loader.add_constructor("!ProviderState", provider_constructor)
    loader.add_constructor("!NodeState", node_constructor)
    loader.add_constructor("!ServiceState", service_constructor)
    loader.add_constructor("!ResourceConfig", resource_config_constructor)
    loader.add_constructor("!ProviderConfig", provider_config_constructor)
    loader.add_constructor("!BaseConfig", base_config_constructor)
    loader.add_constructor("!Config", config_constructor)
    return loader


def get_dumper():
    safe_dumper = yaml.SafeDumper
    safe_dumper.add_representer(NetworkState, network_representer)
    safe_dumper.add_representer(ProviderState, provider_representer)
    safe_dumper.add_representer(NodeState, node_representer)
    safe_dumper.add_representer(ServiceState, service_representer)
    safe_dumper.add_representer(ResourceConfig, resource_config_representer)
    safe_dumper.add_representer(ProviderConfig, provider_config_representer)
    safe_dumper.add_representer(Config, config_representer)
    safe_dumper.add_representer(BaseConfig, base_config_representer)
    safe_dumper.yaml_multi_representers[tuple] = named_tuple
    return safe_dumper



import unittest
from amscrot.client.client import Client
from amscrot.client.job import Job, JobType, JobServiceType, JobSpec, JobState
from amscrot.serviceclient import ServiceClient, PlanError
from amscrot.util.constants import Constants


class TestSessionLifecycle(unittest.TestCase):
    def test_session_job_orchestration(self):
        # 1. Setup Client and Session
        client = Client()

        session = client.create_session("test-session-k8s-v2")
        
        # 2. Setup KubeServiceClient and Job
        k_client = ServiceClient.create(type=Constants.ServiceType.KUBE,
                                        name="sess-k8s")
        
        # 3. Provide Job specification
        spec = JobSpec(
            image="busybox",
            executable=["sleep", "5"],
            resources={"requests": {"cpu": "1", "memory": "1Gi"}},
            attributes={
                "namespace": "default",
                "labels": {"kueue.x-k8s.io/queue-name": "compute-queue"},
                "completions": 5,
                "parallelism": 5,
                "ttlSecondsAfterFinished": 60,
                "priorityClassName": "batch-low",
                "restartPolicy": "Never"
            }
        )
        
        # 4. Add Job with bound spec
        job = Job(
            name="sess-job-1",
            type=JobType.COMPUTE,
            service_type=JobServiceType.BATCH,
            service_client=k_client,
            job_spec=spec
        )
        
        session.add_job(job)
        
        print("\n--- Session Plan ---")
        try:
            session.plan()
        except PlanError as e:
            if any("unreachable" in err for err in e.errors):
                self.skipTest(f"Skipping test - K8s connectivity failed: {e}")
            self.fail(f"Plan failed with validation errors: {e}")
        
        # 6. Apply
        print("\n--- Session Apply ---")
        session.apply()

        # 7. Wait for Completion
        print("\n--- Session Wait ---")
        try:
            results = session.wait(
                jobs=[job],
                target_states=[JobState.COMPLETED, JobState.FAILED, JobState.CANCELED],
                timeout=180.0,
                interval=2.0,
                verbose=True,
            )
            final = results["sess-job-1"]
            print(f"Final Status: {final.state}")
            if final.provider_status:
                succeeded = final.provider_status.get("succeeded", 0)
                print(f"Succeeded count: {succeeded}")
            self.assertEqual(final.state, JobState.COMPLETED, f"Job did not complete: {final}")

        finally:
            # 8. Destroy
            print("\n--- Session Destroy ---")
            session.destroy()

            # Verify cleanup
            final_status = k_client.status(job_name="sess-job-1")
            print(f"Final Status: {final_status.state}")


if __name__ == "__main__":
    unittest.main()



from amscrot.policy.policy_helper import get_stitch_port_for_provider
from amscrot.exceptions import ResourceTypeNotSupported, ProviderException
from amscrot.provider.api.provider import Provider
from amscrot.util.utils import get_logger
from .sense_exceptions import SenseException
from .sense_constants import *

logger = get_logger()


class SenseProvider(Provider):
    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        self.supported_resources = [Constants.RES_TYPE_NETWORK,  Constants.RES_TYPE_NODE]
        self.initialized = False
        self._handled_modify = False

    def setup_environment(self):
        for attr in SENSE_CONF_ATTRS:
            if self.config.get(attr.lower()) is None and self.config.get(attr.upper()) is None:
                raise ProviderException(f"{self.name}: Expecting a value for {attr}")

        pkey = self.config[SENSE_SLICE_PRIVATE_KEY_LOCATION]
        if not pkey:
            return

        from amscrot.util.utils import can_read, is_private_key, absolute_path

        pkey = absolute_path(pkey)

        if not can_read(pkey) or not is_private_key(pkey):
            raise ProviderException(f"{self.name}: unable to read/parse ssh key in {pkey}")

        self.config[SENSE_SLICE_PRIVATE_KEY_LOCATION] = pkey

    @property
    def private_key_file_location(self):
        from .sense_constants import SENSE_SLICE_PRIVATE_KEY_LOCATION

        return self.config.get(SENSE_SLICE_PRIVATE_KEY_LOCATION)

    def _handle_peering_config(self, resource):
        import amscrot.provider.api.dependency_util as util
        from amscrot.model import Network

        peering = resource.get(Constants.RES_PEERING)
        if not peering:
            self.logger.info(f"{self.name}: No peering config found for {resource.get(Constants.LABEL)}")
            return
        prop = 'stitch_interface'
        assert Constants.RES_SECURITY in peering.attributes

        if peering and util.has_resolved_external_dependencies(resource=resource, attribute=prop):
            net = util.get_single_value_for_dependency(resource=resource, attribute=prop)

            if isinstance(net, Network):
                iface = net.interface

                if isinstance(iface, list):
                    iface = iface[0]

                if isinstance(iface, dict) and iface.get("provider") == "fabric":
                    peering.attributes[Constants.RES_ID] = iface[Constants.RES_ID]
                    self.logger.info(f"Added id to peering config:id={iface[Constants.RES_ID]}")

        return peering

    def _init_client(self):
        if not self.initialized:
            self.logger.info(f"{self.name}: Initializing sense client")
            from .sense_client import init_client

            init_client(self.config)
            self.logger.info(f"{self.name}: Initialized sense client")
            self.initialized = True

    def do_add_resource(self, *, resource: dict):
        self._init_client()

        creation_details = resource[Constants.RES_CREATION_DETAILS]

        if not creation_details['in_config_file']:
            return

        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)

        if rtype not in self.supported_resources:
            raise ResourceTypeNotSupported(f"{rtype} for {label}")

        if rtype == Constants.RES_TYPE_NODE:
            from .sense_node import SenseNode
            import amscrot.provider.api.dependency_util as util
            from . import sense_utils

            assert util.has_resolved_internal_dependencies(resource=resource, attribute='network')
            net = util.get_single_value_for_dependency(resource=resource, attribute='network')
            profile_uuid = sense_utils.get_profile_uuid(profile=net.profile)
            vms = sense_utils.get_vms_specs_from_profile(profile_uuid=profile_uuid)

            count = resource[Constants.RES_COUNT]

            if count != len(vms):
                from amscrot.exceptions import ProviderException

                raise ProviderException(f'count {count} != sense_vm_specs: {len(vms)}')

            for idx, vm in enumerate(vms):
                node_name = self.resource_name(resource, idx)
                node = SenseNode(label=label, name=f'{node_name}', network=net.name, spec=vm, provider=self)
                self._nodes.append(node)

                if self.resource_listener:
                    self.resource_listener.on_added(source=self, provider=self, resource=node)

            return

        sense_stitch_port = get_stitch_port_for_provider(resource=resource, provider=self.type)
        interfaces = resource.get(Constants.RES_INTERFACES)

        if not interfaces and sense_stitch_port and 'option' in sense_stitch_port:
            interfaces = sense_stitch_port['option'].get('interface')

        if interfaces:
            for interface in interfaces:
                if not interface.get(Constants.RES_BANDWIDTH):
                    interface[Constants.RES_BANDWIDTH] = resource.get(Constants.RES_BANDWIDTH)

        net_name = self.resource_name(resource)
        profile = resource.get(Constants.RES_PROFILE)

        if not profile and sense_stitch_port:
            profile = sense_stitch_port.get(Constants.RES_PROFILE)

        if not profile:
            raise SenseException(f"Must have a profile for {net_name}")

        layer3 = resource.get(Constants.RES_LAYER3)
        peering = self._handle_peering_config(resource)
        bandwidth = resource.get(Constants.RES_BANDWIDTH)

        from .sense_network import SenseNetwork

        saved_interfaces = self.retrieve_attribute_from_saved_state(resource, net_name, attribute='interface')
        if saved_interfaces and not isinstance(saved_interfaces, list):
            saved_interfaces = [saved_interfaces]

        net = SenseNetwork(label=label, name=net_name, profile=profile,
                           bandwidth=bandwidth, layer3=layer3, peering=peering, interfaces=interfaces,
                           saved_interfaces=saved_interfaces)

        self._networks.append(net)

        if self.resource_listener:
            self.resource_listener.on_added(source=self, provider=self, resource=net)

    def do_create_resource(self, *, resource: dict):
        assert self.initialized
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)
        self.logger.info(f"{self.name}: Creating resource {label} ....")

        if not self._handled_modify and self.modified:
            assert rtype == Constants.RES_TYPE_NETWORK, "sense expects network to be created first"
            self._handled_modify = True

            self.logger.warning(f"{self.name}:Modified state: Deleting sense resources ...")

            from .sense_network import SenseNetwork

            net_name = self.resource_name(resource)
            net = SenseNetwork(label=label, name=net_name, bandwidth=None, profile=None, layer3=None, interfaces=None,
                               peering=None)

            try:
                net.delete()
                self.logger.warning(f"Deleted sense resources ....")
            except Exception as e:
                self.logger.error(f"Exception deleting cloudlab resources ....", e)

        label = resource.get(Constants.LABEL)

        if rtype == Constants.RES_TYPE_NODE:
            for node in [node for node in self._nodes if node.label == label]:
                self.logger.debug(f"Creating node: {vars(node)}")
                node.create()

                if self.resource_listener:
                    self.resource_listener.on_created(source=self, provider=self, resource=node)

                self.logger.debug(f"Created node: {vars(node)}")

            return

        for net in [net for net in self._networks if net.label == label]:
            self.logger.debug(f"Creating network: {vars(net)}")
            net.create()

            if self.resource_listener:
                self.resource_listener.on_created(source=self, provider=self, resource=net)

            self.logger.debug(f"Created network: {vars(net)}")

    def do_delete_resource(self, *, resource: dict):
        self._init_client()
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)

        if rtype == Constants.RES_TYPE_NODE:
            # DO NOTHING
            return

        net_name = self.resource_name(resource)

        logger.debug(f"Deleting network: {net_name}")

        from .sense_network import SenseNetwork

        net = SenseNetwork(label=label, name=net_name, bandwidth=None, profile=None, layer3=None, interfaces=None,
                           peering=None)

        net.delete()
        logger.debug(f"Done Deleting network: {net_name}")

        if self.resource_listener:
            self.resource_listener.on_deleted(source=self, provider=self, resource=net)




from typing import Dict, List, Any, Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from amscrot.serviceclient import ServiceClient
from amscrot.amscrot_manager import AmSCROTManager
from amscrot.util.constants import Constants
from amscrot.util import utils
from .models import Session, Provider

class ProviderCredential:
    def __init__(self, **kwargs):
        self._attributes = kwargs

    def __getattr__(self, item):
        return self._attributes.get(item)

    def __setattr__(self, key, value):
        if key == "_attributes":
            super().__setattr__(key, value)
        else:
            self._attributes[key] = value

    def to_dict(self) -> Dict:
        return self._attributes.copy()

    def update(self, **kwargs):
        self._attributes.update(kwargs)


class Client:
    def __init__(self, *, create_service_clients: bool = False, credential_file: str = None):
        self._providers: List[Provider] = []
        self._sessions: List[Session] = []
        self._service_clients: Dict[str, "ServiceClient"] = {}
        self._credentials = {}
        self._logger = utils.get_logger()

        if create_service_clients:
            self.load_credentials(file_path=credential_file)
            self._create_service_clients_from_credentials()

    def load_credentials(self, *, file_path: str = None):
        """
        Load credentials from a YAML file.
        If file_path is not provided, defaults to ~/.amscrot/credentials.yml.
        """
        from amscrot.util.utils import load_yaml_from_file
        import os

        if file_path is None:
            file_path = "~/.amscrot/credentials.yml"
            path_expanded = os.path.expanduser(file_path)
            if not os.path.exists(path_expanded):
                # Don't overwrite existing credentials if default file is missing
                if not self._credentials:
                    self._credentials = {}
                return

        creds = load_yaml_from_file(file_path) or {}
        # Merge dicts to ProviderCredential objects
        for k, v in creds.items():
            if k in self._credentials:
                self._credentials[k].update(**v)
            else:
                self._credentials[k] = ProviderCredential(**v)

    def add_credential(self, *, profile: str, **kwargs):
        """
        Add or update a credential profile programmatically.
        """
        if profile in self._credentials:
            self._credentials[profile].update(**kwargs)
        else:
            self._credentials[profile] = ProviderCredential(**kwargs)

    def update_credential(self, *, profile: str, **kwargs):
        """
        Update an existing credential profile.
        """
        if profile not in self._credentials:
            raise ValueError(f"Profile '{profile}' not found.")
        self._credentials[profile].update(**kwargs)

    def get_credential(self, profile: str) -> "ProviderCredential":
        """
        Get a specific credential object.
        """
        return self._credentials.get(profile)

    def list_credentials(self) -> List[str]:
        """
        List available credential profiles.
        """
        return list(self._credentials.keys())

    def add_provider(self, *, label: str, type: str, profile: str = None, **kwargs) -> Provider:
        """
        Add a provider configuration.
        """
        # Load credentials if a file is specified
        if "credential_file" in kwargs:
            self.load_credentials(file_path=kwargs["credential_file"])

        attributes = {}

        if profile:
            if profile not in self._credentials:
                raise ValueError(f"Profile '{profile}' not found in loaded credentials.")
            attributes.update(self._credentials[profile].to_dict())
            attributes['profile'] = profile

        attributes.update(kwargs)

        provider = Provider(label, type, **attributes)
        self._providers.append(provider)

        # Forward to any existing sessions so providers can be added after create_session
        for session in self._sessions:
            session.add_provider(provider)

        return provider

    def add_service_client(self, service_client: "ServiceClient"):
        self._service_clients[service_client.name] = service_client
        return service_client

    def get_service_client(self, name: str = None) -> Union[Optional["ServiceClient"], List["ServiceClient"]]:
        """Retrieve a service client by name, or list all if no name given."""
        if name is None:
            return list(self._service_clients.values())
        return self._service_clients.get(name)

    def _create_service_clients_from_credentials(self):
        """Auto-create ServiceClient instances from credential entries that
        contain a ``client_type`` attribute mapping to a valid
        ``Constants.ServiceType`` name (e.g. ``ESNET_IRI``, ``NERSC_IRI``).
        """
        from amscrot.serviceclient import ServiceClient as SC

        for entry_name, cred in self._credentials.items():
            client_type_attr = getattr(cred, 'client_type', None)
            if not client_type_attr:
                continue

            # Resolve e.g. "ESNET_IRI" -> Constants.ServiceType.ESNET_IRI -> "esnet-iri"
            service_type = getattr(Constants.ServiceType, client_type_attr, None)
            if service_type is None:
                self._logger.warning(
                    f"[Client] Unknown client_type '{client_type_attr}' "
                    f"in credential entry '{entry_name}', skipping."
                )
                continue

            try:
                sc = SC.create(
                    type=service_type,
                    name=entry_name,
                    profile=entry_name,
                    credential=cred,
                )
                self._service_clients[entry_name] = sc
                self._logger.debug(
                    f"[Client] Created service client '{entry_name}' "
                    f"(type={service_type})"
                )
            except Exception as e:
                self._logger.warning(
                    f"[Client] Failed to create service client '{entry_name}' "
                    f"(type={service_type}): {e}"
                )

    def create_session(self, name: str) -> Session:
        session = Session(name=name, providers=list(self._providers), service_clients=list(self._service_clients.values()))
        self._sessions.append(session)
        return session



import time
import paramiko
import sys

from amscrot.util.utils import get_logger
from amscrot.util.constants import Constants

logger = get_logger()


class SshNodeTester:
    def __init__(self, *, nodes):
        self.nodes = nodes
        self.helpers = []
        self.passed_ssh_test = []
        self.failed_validation = []
        self.failed_ssh_test = []
        self.summary = ''

        dataplane_addresses = [n.get_dataplane_address(af=Constants.IPv4) for n in self.nodes if
                               n.get_dataplane_address(af=Constants.IPv4)]
        self.run_ping_test = len(dataplane_addresses)

        self.passed_dataplane_ping_tests = {}
        self.failed_dataplane_ping_tests = {}

        if self.run_ping_test:
            self.dataplane_addresses = dataplane_addresses

            for n in self.nodes:
                self.passed_dataplane_ping_tests[n.name] = []

        dataplane_addresses = [n.get_dataplane_address(af=Constants.IPv6) for n in self.nodes if
                               n.get_dataplane_address(af=Constants.IPv6)]
        self.run_ipv6_ping_test = len(dataplane_addresses)

        self.passed_ipv6_dataplane_ping_tests = {}
        self.failed_ipv6_dataplane_ping_tests = {}

        if self.run_ipv6_ping_test:
            self.ipv6_dataplane_addresses = dataplane_addresses

            for n in self.nodes:
                self.passed_ipv6_dataplane_ping_tests[n.name] = []

        for n in self.nodes:
            if self.run_ping_test:
                if not n.get_dataplane_address(af=Constants.IPv4):
                    self.failed_validation.append(dict(node=n.label, description="Missing ipv4 dataplane ip"))
                    continue

            # if self.run_ipv6_ping_test:
            #     if not n.get_dataplane_address(af=Constants.IPv6):
            #         self.failed_validation.append(dict(node=n.label, description="Missing ipv6 dataplane"))
            #         continue

            if n.keyfile is None or n.host is None or n.user is None:
                self.failed_validation.append(
                    dict(node=n.label,
                         description=f"Missing node info:user={n.user}:host={n.host}:keyfile={n.keyfile}"))
                continue

            jump_box = (n.jump_keyfile is None, n.jump_host is None, n.jump_user is None)

            if jump_box != (True, True, True) and jump_box != (False, False, False):
                fmt = "Missing bastion info:user={}:host={}:keyfile={}"
                self.failed_validation.append(
                    dict(node=n.label,
                         description=fmt.format(n.jump_user, n.jump_host, n.jump_keyfile)))
                continue

            helper = SshNodeHelper(node=n, label=n.name, host=n.host,
                                   user=n.user,
                                   private_key_file=n.keyfile,
                                   jump_host=n.jump_host,
                                   jump_user=n.jump_user,
                                   jump_private_key_file=n.jump_keyfile)
            self.helpers.append(helper)

    def has_failures(self):
        return self.failed_validation \
               or self.failed_ssh_test or self.failed_dataplane_ping_tests or self.failed_ipv6_dataplane_ping_tests

    def run_ssh_test(self, *, command='ls -l', retry=5, retry_interval=10):
        for helper in self.helpers:
            logger.info(f"SSH executing {command} on Node:{helper.label}")
            failed_attempts = 0

            for attempt in range(retry):
                try:
                    helper.connect()
                    _, stdout, stderr = helper.client.exec_command(command)
                    exit_code = stdout.channel.recv_exit_status()

                    if exit_code:
                        stdout = str(stdout.read(), 'utf-8').replace('\\n', '\n')
                        stderr = str(stderr.read(), 'utf-8').replace('\\n', '\n')
                        sys.stderr.write(stdout)
                        sys.stderr.write(stderr)
                        raise Exception()

                    self.passed_ssh_test.append(helper.label)
                    logger.info(f"Done with SSH {command} on Node:{helper.label}:attempts={attempt + 1}")
                    break
                except Exception as e:
                    logger.warning(f"SSH test failed:{e}. Node:{helper.label}:attempts={attempt + 1}")
                    time.sleep(retry_interval)
                    failed_attempts += 1
                finally:
                    helper.close_quietly()

            if failed_attempts == retry:
                self.failed_ssh_test.append(helper.label)

    def run_dataplane_test(self, *, command='ping -c 3', retry=3, retry_interval=10):
        for helper in self.helpers:
            node = helper.node
            dataplane_addresses = self.dataplane_addresses

            if len(self.dataplane_addresses) > 1:
               dataplane_addresses = [a for a in self.dataplane_addresses if a != node.get_dataplane_address(af=Constants.IPv4)]

            for attempt in range(retry):
                try:
                    helper.connect()

                    for dataplane_address in dataplane_addresses:
                        if dataplane_address in self.passed_dataplane_ping_tests[helper.label]:
                            continue

                        logger.info(f"SSH executing {command} {dataplane_address} on Node: {helper.label}")

                        _, stdout, stderr = helper.client.exec_command(f"{command} {dataplane_address}")
                        exit_code = stdout.channel.recv_exit_status()

                        if exit_code:
                            stdout = str(stdout.read(), 'utf-8').replace('\\n', '\n')
                            stderr = str(stderr.read(), 'utf-8').replace('\\n', '\n')
                            sys.stderr.write(stdout)
                            sys.stderr.write(stderr)
                            raise Exception(f"pinging {dataplane_address}")

                        self.passed_dataplane_ping_tests[helper.label].append(dataplane_address)

                    logger.info(f"Done with SSH {command} on Node: {helper.label}: attempts={attempt + 1}")
                    break
                except Exception as e:
                    logger.warning(f"SSH ping failed: {e}. Node: {helper.label}:attempts={attempt + 1}")
                    time.sleep(retry_interval)
                finally:
                    helper.close_quietly()

            if len(self.passed_dataplane_ping_tests[helper.label]) != len(dataplane_addresses):
                self.failed_dataplane_ping_tests[helper.label] = list(set(dataplane_addresses).difference(
                    self.passed_dataplane_ping_tests[helper.label]))

    def run_ipv6_dataplane_test(self, *, command='ping6 -c 3', retry=3, retry_interval=10):
        for helper in self.helpers:
            logger.info(f"SSH executing {command} on Node: {helper.label}")
            node = helper.node

            ipv6_dataplane_addresses = self.ipv6_dataplane_addresses

            if len(self.ipv6_dataplane_addresses) > 1:
               ipv6_dataplane_addresses = [a for a in self.ipv6_dataplane_addresses if a != node.get_dataplane_address(af=Constants.IPv6)]

            ipv6_dataplane_addresses = self.ipv6_dataplane_addresses

            if len(self.ipv6_dataplane_addresses) > 1:
               ipv6_dataplane_addresses = [a for a in self.ipv6_dataplane_addresses if a != node.get_dataplane_address(af=Constants.IPv6)]

            for attempt in range(retry):
                try:
                    helper.connect()

                    for dataplane_address in ipv6_dataplane_addresses:
                        if dataplane_address in self.passed_ipv6_dataplane_ping_tests[helper.label]:
                            continue

                        _, stdout, stderr = helper.client.exec_command(f"{command} {dataplane_address}")
                        exit_code = stdout.channel.recv_exit_status()

                        if exit_code:
                            stdout = str(stdout.read(), 'utf-8').replace('\\n', '\n')
                            stderr = str(stderr.read(), 'utf-8').replace('\\n', '\n')
                            sys.stderr.write(stdout)
                            sys.stderr.write(stderr)
                            raise Exception(f"pinging {dataplane_address}")

                        self.passed_ipv6_dataplane_ping_tests[helper.label].append(dataplane_address)

                    logger.info(f"Done with SSH {command} on Node: {helper.label}: attempts={attempt + 1}")
                    break
                except Exception as e:
                    logger.warning(f"SSH ping failed: {e}. Node: {helper.label}:attempts={attempt + 1}")
                    time.sleep(retry_interval)
                finally:
                    helper.close_quietly()

            if len(self.passed_ipv6_dataplane_ping_tests[helper.label]) != len(ipv6_dataplane_addresses):
                self.failed_ipv6_dataplane_ping_tests[helper.label] = list(set(ipv6_dataplane_addresses).difference(
                    self.passed_ipv6_dataplane_ping_tests[helper.label]))

    def run_tests(self, *, retry=3, retry_interval=10):
        from collections import namedtuple

        # self.run_ssh_test(retry=retry, retry_interval=retry_interval)

        if self.run_ping_test:
            self.run_dataplane_test(retry=retry, retry_interval=retry_interval)
        else:
            self.run_ssh_test(retry=retry, retry_interval=retry_interval)

        # if self.run_ipv6_ping_test:
        #     self.run_ipv6_dataplane_test(retry=retry, retry_interval=retry_interval)

        Node = namedtuple("Node", "host user key_file data_plane_address")
        JumpNode = namedtuple("Node", "host user key_file data_plane_address jump_host jump_user jump_key_file")

        node_info_list = []

        for n in self.nodes:
            if n.jump_user or n.jump_host or n.jump_keyfile:
                node_info_list.append(
                    {n.name: JumpNode(host=n.host,
                                       user=n.user,
                                       key_file=n.keyfile,
                                       data_plane_address=n.get_dataplane_address(),
                                       jump_host=n.jump_host,
                                       jump_user=n.jump_user,
                                       jump_key_file=n.jump_keyfile
                                       )
                     }
                )
            else:
                node_info_list.append({n.name: Node(host=n.host,
                                                     user=n.user,
                                                     key_file=n.keyfile,
                                                     data_plane_address=n.get_dataplane_address())})

        if self.run_ping_test or self.run_ipv6_ping_test:
            self.summary = {"SSH TEST SUMMARY": [
                {"nodes": node_info_list},
                {"passed_tests":
                    [
                        # {"passed_ssh_test": self.passed_ssh_test},
                        {"passed_ipv4_dataplane_ping_test": self.passed_dataplane_ping_tests}
                    ]},
                {"FAILED_VALIDATION": self.failed_validation},
                {"FAILED_TESTS":
                    [
                        # {"failed_ssh_test": self.failed_ssh_test},
                        {"failed_ipv4_dataplane_ping_test": [dict(src=k, destinations=v) for k, v in
                                                        self.failed_dataplane_ping_tests.items()]}
                    ]}
            ]}
        else:
            self.summary = {"SSH TEST SUMMARY": [
                {"nodes": node_info_list},
                {"passed_tests":
                    [
                        {"passed_ssh_test": self.passed_ssh_test}
                    ]},
                {"FAILED_VALIDATION": self.failed_validation},
                {"FAILED_TESTS":
                    [
                        {"failed_ssh_test": self.failed_ssh_test}
                    ]}
            ]}


class SshNodeHelper:
    # noinspection PyBroadException
    def __init__(self, *, node, label, host, user, private_key_file, jump_host=None, jump_user=None,
                 jump_private_key_file=None):
        self.node = node
        self.label = label
        self.host = host
        self.user = user
        self.jump_host = jump_host
        self.jump_user = jump_user
        self.client = None
        self.jump_client = None

        # noinspection PyBroadException
        try:
            self.key = paramiko.RSAKey.from_private_key_file(private_key_file)
        except Exception:
            self.key = paramiko.ecdsakey.ECDSAKey.from_private_key_file(private_key_file)

        if jump_private_key_file:
            try:
                self.jump_key = paramiko.RSAKey.from_private_key_file(jump_private_key_file)
            except Exception:
                self.jump_key = paramiko.ecdsakey.ECDSAKey.from_private_key_file(jump_private_key_file)

    def connect(self):
        if not self.jump_user:
            self.client = paramiko.SSHClient()
            # self.client.load_system_host_keys()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.connect(self.host, username=self.user, pkey=self.key)
            return

        self.jump_client = paramiko.SSHClient()
        # self.jump_client.load_system_host_keys()
        self.jump_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.jump_client.connect(self.jump_host, username=self.jump_user, pkey=self.jump_key)

        transport = self.jump_client.get_transport()
        src_addr = (self.jump_host, 22)
        dest_addr = (self.host, 22)
        channel = transport.open_channel("direct-tcpip", dest_addr, src_addr)
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(self.host, username=self.user, pkey=self.key, sock=channel)

    # noinspection PyBroadException
    def close_quietly(self):
        if self.client:
            try:
                self.client.close()
            except Exception:
                pass

            self.client = None

        if self.jump_client:
            try:
                self.jump_client.close()
            except Exception:
                pass

            self.jump_client = None



import unittest
from amscrot.serviceclient import ServiceClient, PlanError
from amscrot.util.constants import Constants
from amscrot.client.job import JobSpec, JobState

class TestServiceClientMethods(unittest.TestCase):
    def test_iri_methods(self):
        client = ServiceClient.create(type=Constants.ServiceType.IRI, name="iri1", endpoint_uri="http://iri")
        spec = JobSpec(image="test-image")
        
        # Test Plan
        plan_result = client.plan(spec)
        self.assertIsInstance(plan_result, dict)
        self.assertEqual(plan_result["status"], JobState.PLANNED)
        
        # Test Create
        client.create(spec)
        self.assertEqual(client.status().state, JobState.ACTIVE)
        
        # Test Destroy
        client.destroy()
        self.assertEqual(client.status().state, JobState.CANCELED)
        
    def test_kube_methods(self):
        client = ServiceClient.create(type=Constants.ServiceType.KUBE, name="kube1", endpoint_uri="http://kube")
        spec = JobSpec(image="busybox", executable=["echo", "hello"])

        try:
            plan_result = client.plan(spec)
        except PlanError as e:
            if any("unreachable" in err for err in e.errors):
                self.skipTest(f"Skipping test - K8s connectivity failed: {e}")
            self.fail(f"Plan raised PlanError unexpectedly: {e}")
        self.assertIsInstance(plan_result, dict)
        self.assertEqual(plan_result["status"], JobState.PLANNED)

        client.create(spec)
        client.destroy()
        client.status()

    def test_kube_log_capture(self):
        # Test with a job that sleeps and prints
        client = ServiceClient.create(type=Constants.ServiceType.KUBE, name="logtest", endpoint_uri="http://kube")
        spec = JobSpec(
            image="python:3.9-slim", 
            executable=["python", "-c", "import time; print('Hello K8s Logs'); time.sleep(5); print('Done Sleep')"]
        )
        
        try:
            plan_result = client.plan(spec)
        except PlanError as e:
            if any("unreachable" in err for err in e.errors):
                self.skipTest(f"Skipping test - K8s connectivity failed: {e}")
            self.fail(f"Plan raised PlanError unexpectedly: {e}")
        self.assertEqual(plan_result["status"], JobState.PLANNED)
        
        client.create(spec)
        
        # Poll for logs (simulated simple polling loop)
        import time
        max_retries = 30
        found_log = False
        
        try:
            for _ in range(max_retries):
                job_status = client.status()
                logs = (job_status.provider_status or {}).get("logs", "")
                if "Hello K8s Logs" in logs:
                    found_log = True
                    print(logs)
                    break
                time.sleep(1)
        finally:
            client.destroy()

        if client._available:
             self.assertTrue(found_log, "Did not find expected log message")
        
if __name__ == "__main__":
    unittest.main()



from collections import namedtuple

Duration = namedtuple("Duration", "duration comment")

Stages = namedtuple("Stages", "setup_duration plan_duration create_duration delete_duration")

ProviderStats = namedtuple("ProviderStats",
                           "provider provider_duration has_failures has_pending stages")

AmSCROTStats = namedtuple("AmSCROTStats",
                          "action has_failures workflow_duration workflow_config controller providers provider_stats")



from abc import ABC, abstractmethod
from typing import List, Any, Dict, Optional, TYPE_CHECKING
from amscrot.util import utils

if TYPE_CHECKING:
    from amscrot.client.job import JobSpec
    from amscrot.model.discovery import DiscoveryResult


class PlanError(Exception):
    """Raised by ServiceClient.plan() when validation errors prevent job submission.

    Attributes:
        errors:   List of error messages that caused the failure.
        warnings: List of non-fatal warning messages (may be empty).
    """

    def __init__(self, errors: List[str], warnings: List[str] = None):
        self.errors = errors
        self.warnings = warnings or []
        error_summary = "; ".join(errors)
        super().__init__(f"Plan failed with {len(errors)} error(s): {error_summary}")


class ServiceClient(ABC):
    def __init__(self, name: str, type: str, endpoint_uri: Optional[str] = None, 
                 status: str = "ACTIVE", capabilities: List[Any] = None, 
                 allocated: List[Any] = None,
                 credential: Any = None, profile: str = None, credential_file: str = None):
        self.name = name
        self.endpoint_uri = endpoint_uri
        self.type = type
        self._status = status
        self.capabilities = capabilities or []
        self.allocated = allocated or []
        self.credential = credential
        self.profile = profile
        self.credential_file = credential_file
        self.logger = utils.get_logger()

    @abstractmethod
    def discover(self, native: bool = True) -> "DiscoveryResult":
        pass

    @abstractmethod
    def plan(self, job_spec: "JobSpec", job_name: str = None) -> Dict:
        pass

    @abstractmethod
    def create(self, job_spec: "JobSpec", job_name: str = None):
        pass

    @abstractmethod
    def destroy(self, job_name: str = None):
        pass

    @abstractmethod
    def status(self, job_name: str = None) -> "JobStatus":
        pass

    @classmethod
    def create(cls, type: str, **kwargs) -> "ServiceClient":
        from amscrot.util.constants import Constants
        import importlib

        class_path = Constants.SERVICE_CLIENT_CLASSES.get(type)
        if not class_path:
             raise ValueError(f"Unknown ServiceClient type: {type}")

        try:
            module_name, class_name = class_path.rsplit('.', 1)
            module = importlib.import_module(module_name)
            client_class = getattr(module, class_name)
            return client_class(**kwargs)
        except (ImportError, AttributeError) as e:
            raise ImportError(f"Failed to load ServiceClient class for type '{type}' from '{class_path}': {e}")

    def to_config(self) -> Dict:
        return {
            'service_client': [
                {
                    self.name: {
                        'type': self.type,
                        'endpoint_uri': self.endpoint_uri,
                        'status': self._status
                    }
                }
            ]
        }

    def __repr__(self):
        return f"<ServiceClient name={self.name} type={self.type} uri={self.endpoint_uri}>"